import json
from pathlib import Path
from unittest.mock import patch

import pytest
from dodal.devices.aperturescatterguard import ApertureValue
from pydantic import ValidationError

from mx_bluesky.common.external_interaction.callbacks.common.grid_detection_callback import (
    GridParamUpdate,
)
from mx_bluesky.common.parameters.constants import GridscanParamConstants
from mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan import (
    create_parameters_for_flyscan_xray_centre,
)
from mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan import (
    create_parameters_for_grid_detection,
)
from mx_bluesky.hyperion.parameters.gridscan import (
    HyperionSpecifiedThreeDGridScan,
    OddYStepsException,
)
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect
from mx_bluesky.hyperion.parameters.robot_load import RobotLoadThenCentre
from mx_bluesky.hyperion.parameters.rotation import RotationScan

from ....conftest import raw_params_from_file


@pytest.fixture
def load_centre_collect_params_with_panda():
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_load_centre_collect_params.json"
    )
    params["robot_load_then_centre"]["features"]["use_panda_for_gridscan"] = True
    return LoadCentreCollect(**params)


@pytest.fixture
def minimal_3d_gridscan_params():
    return {
        "sample_id": 123,
        "x_start_um": 0.123,
        "y_start_um": 0.777,
        "z_start_um": 0.05,
        "parameter_model_version": "5.0.0",
        "visit": "cm12345",
        "file_name": "test_file_name",
        "y2_start_um": 2,
        "z2_start_um": 2,
        "x_steps": 5,
        "y_steps": 7,
        "z_steps": 9,
        "storage_directory": "/tmp/dls/i03/data/2024/cm31105-4/xraycentring/123456/",
    }


def get_empty_grid_parameters() -> GridParamUpdate:
    return {
        "x_start_um": 1,
        "y_start_um": 1,
        "y2_start_um": 1,
        "z_start_um": 1,
        "z2_start_um": 1,
        "x_steps": 1,
        "y_steps": 1,
        "z_steps": 1,
        "x_step_size_um": 1,
        "y_step_size_um": 1,
        "z_step_size_um": 1,
    }


def test_minimal_3d_gridscan_params(minimal_3d_gridscan_params):
    test_params = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    assert {"sam_x", "sam_y", "sam_z"} == set(test_params.scan_points.keys())
    assert test_params.scan_indices == [0, 35]
    assert test_params.num_images == (5 * 7 + 5 * 9)
    assert test_params.exposure_time_s == GridscanParamConstants.EXPOSURE_TIME_S


def test_cant_do_panda_fgs_with_odd_y_steps(minimal_3d_gridscan_params):
    test_params = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    with pytest.raises(OddYStepsException):
        _ = test_params.panda_FGS_params
    assert test_params.FGS_params


def test_serialise_deserialise(minimal_3d_gridscan_params):
    test_params = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    serialised = json.loads(test_params.model_dump_json())
    deserialised = HyperionSpecifiedThreeDGridScan(**serialised)
    assert deserialised.demand_energy_ev is None
    assert deserialised.visit == "cm12345"
    assert deserialised.x_start_um == 0.123


@pytest.mark.parametrize(
    "version, valid",
    [
        ("4.3.0", False),
        ("6.3.7", False),
        ("5.0.0", True),
        ("5.3.0", True),
        ("5.3.7", True),
    ],
)
def test_param_version(minimal_3d_gridscan_params, version: str, valid: bool):
    minimal_3d_gridscan_params["parameter_model_version"] = version
    if valid:
        _ = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    else:
        with pytest.raises(ValidationError):
            _ = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)


def test_robot_load_then_centre_params():
    params = {
        "parameter_model_version": "5.0.0",
        "sample_id": 123456,
        "visit": "cm12345",
        "file_name": "file_name",
        "storage_directory": "/tmp/dls/i03/data/2024/cm31105-4/xraycentring/123456/",
    }
    params["detector_distance_mm"] = 200
    test_params = RobotLoadThenCentre(**params)
    assert test_params.detector_params


def test_default_snapshot_path(minimal_3d_gridscan_params):
    gridscan_params = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    assert gridscan_params.snapshot_directory == Path(
        "/tmp/dls/i03/data/2024/cm31105-4/xraycentring/123456/snapshots"
    )

    params_with_snapshot_path = dict(minimal_3d_gridscan_params)
    params_with_snapshot_path["snapshot_directory"] = "/tmp/my_snapshots"

    gridscan_params_with_snapshot_path = HyperionSpecifiedThreeDGridScan(
        **params_with_snapshot_path
    )
    assert gridscan_params_with_snapshot_path.snapshot_directory == Path(
        "/tmp/my_snapshots"
    )


def test_osc_is_used():
    raw_params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_rotation_scan_parameters.json"
    )
    for osc in [0.001, 0.05, 0.1, 0.2, 0.75, 1, 1.43]:
        raw_params["rotation_increment_deg"] = osc
        params = RotationScan(**raw_params)
        assert params.rotation_increment_deg == osc
        assert params.num_images == int(params.scan_width_deg / osc)


def test_selected_aperture_uses_default():
    raw_params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_rotation_scan_parameters.json"
    )
    raw_params["selected_aperture"] = None
    params = RotationScan(**raw_params)
    assert params.selected_aperture == ApertureValue.LARGE


def test_feature_flags_overriden_if_supplied(minimal_3d_gridscan_params):
    test_params = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    assert test_params.features.use_panda_for_gridscan is False
    assert test_params.features.compare_cpu_and_gpu_zocalo is False
    assert test_params.features.use_gpu_results is False
    minimal_3d_gridscan_params["features"] = {
        "use_panda_for_gridscan": True,
        "compare_cpu_and_gpu_zocalo": True,
    }
    test_params = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    assert test_params.features.compare_cpu_and_gpu_zocalo
    assert test_params.features.use_panda_for_gridscan
    # Config server shouldn't update values which were explicitly provided
    test_params.features.update_self_from_server()
    assert test_params.features.compare_cpu_and_gpu_zocalo
    assert test_params.features.use_panda_for_gridscan


@pytest.mark.parametrize(
    "feature_set, expected_dev_shm",
    [
        (
            {
                "compare_cpu_and_gpu_zocalo": True,
                "use_gpu_results": False,
            },
            True,
        ),
        (
            {
                "compare_cpu_and_gpu_zocalo": False,
                "use_gpu_results": True,
            },
            True,
        ),
        (
            {
                "compare_cpu_and_gpu_zocalo": False,
                "use_gpu_results": False,
            },
            False,
        ),
    ],
)
@patch("mx_bluesky.common.parameters.components.os")
def test_gpu_enabled_if_use_gpu_results_or_compare_gpu_enabled(
    _, feature_set, expected_dev_shm, minimal_3d_gridscan_params
):
    minimal_3d_gridscan_params["detector_distance_mm"] = 100

    grid_scan = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    assert not grid_scan.detector_params.enable_dev_shm

    minimal_3d_gridscan_params["features"] = feature_set
    grid_scan = HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)
    assert grid_scan.detector_params.enable_dev_shm == expected_dev_shm


@patch("mx_bluesky.common.parameters.components.os")
def test_if_use_gpu_results_and_compare_gpu_enabled_then_validation_error(
    _, minimal_3d_gridscan_params
):
    minimal_3d_gridscan_params["features"] = {
        "compare_cpu_and_gpu_zocalo": True,
        "use_gpu_results": True,
    }
    with pytest.raises(ValidationError):
        HyperionSpecifiedThreeDGridScan(**minimal_3d_gridscan_params)


def test_hyperion_params_correctly_carried_through_UDC_parameter_models(
    load_centre_collect_params_with_panda: LoadCentreCollect,
):
    robot_load_then_centre_params = (
        load_centre_collect_params_with_panda.robot_load_then_centre
    )
    assert robot_load_then_centre_params.detector_params.enable_dev_shm
    pin_tip_then_xrc_params = (
        robot_load_then_centre_params.pin_centre_then_xray_centre_params
    )
    assert pin_tip_then_xrc_params.detector_params.enable_dev_shm
    grid_detect_then_xrc_params = create_parameters_for_grid_detection(
        pin_tip_then_xrc_params
    )
    assert pin_tip_then_xrc_params.detector_params.enable_dev_shm
    flyscan_xrc_params = create_parameters_for_flyscan_xray_centre(
        grid_detect_then_xrc_params, get_empty_grid_parameters()
    )
    assert flyscan_xrc_params.detector_params.enable_dev_shm
    assert flyscan_xrc_params.panda_runup_distance_mm == 0.17
    assert flyscan_xrc_params.features.use_panda_for_gridscan
    assert flyscan_xrc_params.features.compare_cpu_and_gpu_zocalo



from __future__ import annotations

import os
import re
import signal
import subprocess
import threading
from genericpath import isfile
from time import sleep
from unittest.mock import MagicMock, patch

import bluesky.plan_stubs as bps
import pytest
import zmq
from bluesky.callbacks import CallbackBase
from bluesky.callbacks.zmq import Publisher
from bluesky.run_engine import RunEngine
from dodal.devices.zocalo.zocalo_results import (
    get_processing_results_from_event,
)
from zmq.utils.monitor import recv_monitor_message

from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.utils.utils import convert_angstrom_to_eV
from mx_bluesky.hyperion.experiment_plans.flyscan_xray_centre_plan import (
    flyscan_xray_centre,
)
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    rotation_scan,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionFlyScanXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan
from mx_bluesky.hyperion.parameters.rotation import RotationScan

from .....conftest import TEST_RESULT_LARGE, fake_read
from ..conftest import (  # noqa
    fetch_comment,
)

"""
Note that because these tests use the external processes some of the errors coming from
them may not be very informative. You will want to check the log files produced in `tmp`
for better logs.
"""


class DocumentCatcher(CallbackBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start = MagicMock()
        self.descriptor = MagicMock()
        self.event = MagicMock()
        self.stop = MagicMock()


def event_monitor(monitor: zmq.Socket, connection_active_lock: threading.Lock) -> None:
    try:
        while monitor.poll():
            monitor_event = recv_monitor_message(monitor)
            LOGGER.info(f"Event: {monitor_event}")
            if monitor_event["event"] == zmq.EVENT_CONNECTED:
                LOGGER.info("CONNECTED - acquiring connection_active_lock")
                connection_active_lock.acquire()
            if monitor_event["event"] == zmq.EVENT_MONITOR_STOPPED:
                break
    except zmq.ZMQError:
        pass
    finally:
        connection_active_lock.release()
        monitor.close()
        LOGGER.info("event monitor thread done!")


@pytest.fixture
def RE_with_external_callbacks(
    zocalo_env,  # ZOCALO_CONFIG must be exported to external callback environment
):
    RE = RunEngine()

    process_env = os.environ.copy()

    external_callbacks_process = subprocess.Popen(
        [
            "python",
            "src/mx_bluesky/hyperion/external_interaction/callbacks/__main__.py",
            "--dev",
        ],
        env=process_env,
    )
    publisher = Publisher(f"localhost:{CONST.CALLBACK_0MQ_PROXY_PORTS[0]}")
    monitor = publisher._socket.get_monitor_socket()

    connection_active_lock = threading.Lock()
    t = threading.Thread(
        target=event_monitor,
        args=(monitor, connection_active_lock),
        name="event_monitor",
    )
    t.start()

    while not connection_active_lock.locked():
        sleep(0.1)  # wait for connection to happen before continuing

    sub_id = RE.subscribe(publisher)

    yield RE

    RE.unsubscribe(sub_id)
    publisher.close()

    external_callbacks_process.send_signal(signal.SIGINT)
    sleep(0.01)
    external_callbacks_process.kill()
    external_callbacks_process.wait(10)
    t.join()


@pytest.mark.system_test
def test_RE_with_external_callbacks_starts_and_stops(
    RE_with_external_callbacks: RunEngine,
):
    RE = RE_with_external_callbacks

    def plan():
        yield from bps.sleep(1)

    RE(plan())


@pytest.mark.system_test
async def test_external_callbacks_handle_gridscan_ispyb_and_zocalo(
    RE_with_external_callbacks: RunEngine,
    dummy_params: HyperionSpecifiedThreeDGridScan,
    fgs_composite_for_fake_zocalo: HyperionFlyScanXRayCentreComposite,
    done_status,
    fetch_comment,  # noqa
):
    """
    This test requires fake zocalo, and a connection to the dev ISPyB database.
    """

    RE = RE_with_external_callbacks

    doc_catcher = DocumentCatcher()
    RE.subscribe(doc_catcher)

    # Run the xray centring plan
    RE(flyscan_xray_centre(fgs_composite_for_fake_zocalo, dummy_params))

    # Check that we we emitted a valid reading from the zocalo device
    zocalo_event = doc_catcher.event.call_args.args[0]  # type: ignore
    # TEST_RESULT_LARGE is what fake_zocalo sends by default
    assert (
        get_processing_results_from_event("zocalo", zocalo_event) == TEST_RESULT_LARGE
    )

    # get dcids from zocalo device
    dcid_reading = await fgs_composite_for_fake_zocalo.zocalo.ispyb_dcid.read()
    dcgid_reading = await fgs_composite_for_fake_zocalo.zocalo.ispyb_dcgid.read()

    dcid = dcid_reading["zocalo-ispyb_dcid"]["value"]
    dcgid = dcgid_reading["zocalo-ispyb_dcgid"]["value"]

    assert dcid != 0
    assert dcgid != 0

    # check the data in dev ispyb corresponding to this "collection"
    ispyb_comment = fetch_comment(dcid)
    assert ispyb_comment != ""
    assert "Zocalo processing took" in ispyb_comment
    assert "Position (grid boxes) ['1.0', '2.0', '3.0']" in ispyb_comment
    assert "Size (grid boxes) [6 6 5];" in ispyb_comment


@pytest.mark.system_test
def test_remote_callbacks_write_to_dev_ispyb_for_rotation(
    RE_with_external_callbacks: RunEngine,
    params_for_rotation_scan: RotationScan,
    fetch_comment,  # noqa
    fetch_datacollection_attribute,
    composite_for_rotation_scan,
    oav_parameters_for_rotation,
):
    test_wl = 0.71
    test_bs_x = 0.020
    test_bs_y = 0.020
    test_exp_time = 0.023
    test_img_wid = 0.27

    params_for_rotation_scan.rotation_increment_deg = test_img_wid
    params_for_rotation_scan.exposure_time_s = test_exp_time
    params_for_rotation_scan.demand_energy_ev = convert_angstrom_to_eV(test_wl)

    with patch("bluesky.preprocessors.__read_and_stash_a_motor", fake_read):
        RE_with_external_callbacks(
            rotation_scan(
                composite_for_rotation_scan,
                params_for_rotation_scan,
                oav_parameters_for_rotation,
            )
        )

    sleep(1)
    assert isfile("/tmp/logs/bluesky/hyperion_ispyb_callback.log")
    ispyb_log_tail = subprocess.run(
        ["tail", "/tmp/logs/bluesky/hyperion_ispyb_callback.log"],
        timeout=1,
        stdout=subprocess.PIPE,
    ).stdout.decode("utf-8")

    ids_re = re.compile(
        r"data_collection_ids=\((\d+),\) data_collection_group_id=(\d+) "
    )
    matches = ids_re.findall(ispyb_log_tail)

    dcid = matches[0][0]

    comment = fetch_comment(dcid)
    assert comment == "test Sample position (µm): (1, 2, 3) Aperture: Small. "
    wavelength = fetch_datacollection_attribute(dcid, "wavelength")
    beamsize_x = fetch_datacollection_attribute(dcid, "beamSizeAtSampleX")
    beamsize_y = fetch_datacollection_attribute(dcid, "beamSizeAtSampleY")
    exposure = fetch_datacollection_attribute(dcid, "exposureTime")

    assert wavelength == test_wl
    assert beamsize_x == test_bs_x
    assert beamsize_y == test_bs_y
    assert exposure == test_exp_time



import pytest
from bluesky.plan_stubs import null

from mx_bluesky.common.utils.exceptions import (
    WarningException,
    catch_exception_and_warn,
)


class _TestException(Exception):
    pass


def dummy_plan():
    yield from null()
    raise _TestException


def test_catch_exception_and_warn_correctly_raises_warning_exception(RE):
    with pytest.raises(WarningException):
        RE(catch_exception_and_warn(_TestException, dummy_plan))


def test_catch_exception_and_warn_correctly_raises_original_exception(RE):
    with pytest.raises(_TestException):
        RE(catch_exception_and_warn(ValueError, dummy_plan))



import math

import pytest

from mx_bluesky.common.utils.utils import (
    convert_angstrom_to_eV,
    convert_eV_to_angstrom,
    energy_to_bragg_angle,
)

SI_111_SPACING_ANGSTROMS = 3.13475

test_wavelengths = [1.620709, 1.2398425, 0.9762539, 0.8265616, 0.68880138]
test_energies = [7650, 10000, 12700, 15000, 18000]


@pytest.mark.parametrize(
    "test_wavelength, test_energy",
    list(zip(test_wavelengths, test_energies, strict=False)),
)
def test_ev_to_a_converter(test_wavelength, test_energy):
    assert convert_eV_to_angstrom(test_energy) == pytest.approx(test_wavelength)


@pytest.mark.parametrize(
    "test_wavelength, test_energy",
    list(zip(test_wavelengths, test_energies, strict=False)),
)
def test_a_to_ev_converter(test_wavelength, test_energy):
    assert convert_angstrom_to_eV(test_wavelength) == pytest.approx(test_energy)


@pytest.mark.parametrize(
    "input_energy_kev, expected_bragg_angle_deg",
    [[7, 16.41], [8, 14.31], [11, 10.35], [12.3, 9.25], [15, 7.57]],
)
def test_energy_to_bragg_angle(
    input_energy_kev: float, expected_bragg_angle_deg: float
):
    assert math.isclose(
        energy_to_bragg_angle(input_energy_kev, SI_111_SPACING_ANGSTROMS),
        expected_bragg_angle_deg,
        abs_tol=0.01,
    )



from unittest.mock import MagicMock

import bluesky.preprocessors as bpp
import pytest
from bluesky import plan_stubs as bps
from bluesky.run_engine import RunEngine
from bluesky.utils import FailedStatus
from dodal.devices.xbpm_feedback import Pause
from dodal.plans.preprocessors.verify_undulator_gap import (
    verify_undulator_gap_before_run_decorator,
)
from ophyd.status import Status
from ophyd_async.testing import set_mock_value
from tests.conftest import XBPMAndTransmissionWrapperComposite

from mx_bluesky.common.preprocessors.preprocessors import (
    transmission_and_xbpm_feedback_for_collection_decorator,
)


@pytest.fixture
def composite(
    xbpm_and_transmission_wrapper_composite: XBPMAndTransmissionWrapperComposite,
    done_status,
) -> XBPMAndTransmissionWrapperComposite:
    xbpm_and_transmission_wrapper_composite.undulator.set = MagicMock(
        return_value=done_status
    )

    return xbpm_and_transmission_wrapper_composite


async def test_xbpm_decorator_with_undulator_check_decorators(
    RE, composite: XBPMAndTransmissionWrapperComposite
):
    energy_in_kev = 11.3
    composite.dcm.energy_in_kev.user_readback.read = MagicMock(
        return_value={"value": {"value": energy_in_kev}}
    )

    @transmission_and_xbpm_feedback_for_collection_decorator(composite, 0.1)
    @verify_undulator_gap_before_run_decorator(composite)
    @bpp.run_decorator()
    def my_collection_plan():
        yield from bps.null()

    set_mock_value(composite.xbpm_feedback.pos_stable, 1)
    RE = RunEngine()
    RE(my_collection_plan())

    # Stop pyright from complaining
    assert isinstance(composite.xbpm_feedback.trigger, MagicMock)
    assert isinstance(composite.undulator.set, MagicMock)

    # Assert XBPM is stable
    composite.xbpm_feedback.trigger.assert_called_once()
    # Assert DCM energy is read after XBPM is stable
    composite.dcm.energy_in_kev.user_readback.read.assert_called_once()
    # Assert Undulator is finally set
    composite.undulator.set.assert_called_once()
    # Assert energy passed to the Undulator is the same as read from the DCM
    assert composite.undulator.set.call_args.args[0] == energy_in_kev


async def test_given_xpbm_checks_pass_when_plan_run_with_decorator_then_run_as_expected(
    RE, composite: XBPMAndTransmissionWrapperComposite
):
    expected_transmission = 0.3

    @transmission_and_xbpm_feedback_for_collection_decorator(
        composite, expected_transmission
    )
    @bpp.run_decorator()
    def my_collection_plan():
        read_transmission = yield from bps.rd(composite.attenuator.actual_transmission)
        assert read_transmission == expected_transmission
        pause_feedback = yield from bps.rd(composite.xbpm_feedback.pause_feedback)
        assert pause_feedback == Pause.PAUSE

    set_mock_value(composite.xbpm_feedback.pos_stable, 1)

    RE = RunEngine()
    RE(my_collection_plan())

    assert await composite.attenuator.actual_transmission.get_value() == 1.0
    assert await composite.xbpm_feedback.pause_feedback.get_value() == Pause.RUN


async def test_given_xbpm_checks_fail_when_plan_run_with_decorator_then_plan_not_run(
    RE, composite: XBPMAndTransmissionWrapperComposite
):
    mock = MagicMock()

    @transmission_and_xbpm_feedback_for_collection_decorator(composite, 0.1)
    @bpp.run_decorator()
    def my_collection_plan():
        mock()
        yield from bps.null()

    status = Status()
    status.set_exception(Exception())
    composite.xbpm_feedback.trigger = MagicMock(side_effect=lambda: status)

    RE = RunEngine()
    with pytest.raises(FailedStatus):
        RE(my_collection_plan())

    mock.assert_not_called()
    assert await composite.attenuator.actual_transmission.get_value() == 1.0
    assert await composite.xbpm_feedback.pause_feedback.get_value() == Pause.RUN


async def test_given_xpbm_checks_pass_and_plan_fails_when_plan_run_with_decorator_then_cleaned_up(
    RE, composite: XBPMAndTransmissionWrapperComposite
):
    set_mock_value(composite.xbpm_feedback.pos_stable, 1)

    class MyException(Exception):
        pass

    @transmission_and_xbpm_feedback_for_collection_decorator(composite, 0.1)
    @bpp.run_decorator()
    def my_collection_plan():
        yield from bps.null()
        raise MyException()

    RE = RunEngine()
    with pytest.raises(MyException):
        RE(my_collection_plan())

    assert await composite.attenuator.actual_transmission.get_value() == 1.0
    assert await composite.xbpm_feedback.pause_feedback.get_value() == Pause.RUN



import os
import pathlib
import pprint
import time
from datetime import datetime

import bluesky.plan_stubs as bps
import requests

from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import ChipType, MappingType
from mx_bluesky.beamlines.i24.serial.log import SSX_LOGGER
from mx_bluesky.beamlines.i24.serial.parameters import (
    ExtruderParameters,
    FixedTargetParameters,
)
from mx_bluesky.beamlines.i24.serial.setup_beamline import Eiger, caget, cagetstring


def call_nexgen(
    chip_prog_dict: dict | None,
    parameters: ExtruderParameters | FixedTargetParameters,
    wavelength_in_a: float,
    beam_center_in_pix: tuple[float, float],
    start_time: datetime | None = None,
):
    """Call the nexus writer by sending a request to nexgen-server.

    Args:
        chip_prog_dict (dict | None): Dictionary containing most of the information \
            passed to the program runner for the collection. Only used for fixed target.
        start_time
        parameters (SerialAndLaserExperiment): Collection parameters.
        wavelength_in_a (float): Wavelength, in A.
        beam_center_in_pix (list[float]): Beam center position on detector, in pixels.
        start_time (datetime, optional): Collection start time.

    Raises:
        ValueError: For a wrong experiment type passed (either unknwon or not matched \
            to parameter model).

    """
    current_chip_map = None
    match parameters:
        case FixedTargetParameters():
            if not (
                parameters.map_type == MappingType.NoMap
                or parameters.chip.chip_type == ChipType.Custom
            ):
                # For nexgen >= 0.9.10
                current_chip_map = parameters.chip_map
            pump_status = bool(parameters.pump_repeat)
            total_numb_imgs = parameters.total_num_images
        case ExtruderParameters():
            total_numb_imgs = parameters.num_images
            pump_status = parameters.pump_status

    filename_prefix = cagetstring(Eiger.pv.filenameRBV)
    meta_h5 = parameters.visit / parameters.directory / f"{filename_prefix}_meta.h5"
    t0 = time.time()
    max_wait = 60  # seconds
    SSX_LOGGER.info(f"Watching for {meta_h5}")
    while time.time() - t0 < max_wait:
        if meta_h5.exists():
            SSX_LOGGER.info(f"Found {meta_h5} after {time.time() - t0:.1f} seconds")
            yield from bps.sleep(5)
            break
        SSX_LOGGER.debug(f"Waiting for {meta_h5}")
        yield from bps.sleep(1)
    if not meta_h5.exists():
        SSX_LOGGER.warning(f"Giving up waiting for {meta_h5} after {max_wait} seconds")
        return

    bit_depth = int(caget(Eiger.pv.bit_depth))
    SSX_LOGGER.debug(
        f"Call to nexgen server with the following chip definition: \n{chip_prog_dict}"
    )

    access_token = pathlib.Path("/scratch/ssx_nexgen.key").read_text().strip()
    url = "https://ssx-nexgen.diamond.ac.uk/ssx_eiger/write"
    headers = {"Authorization": f"Bearer {access_token}"}

    payload = {
        "beamline": "i24",
        "beam_center": beam_center_in_pix,
        "chipmap": current_chip_map,
        "chip_info": chip_prog_dict,
        "det_dist": parameters.detector_distance_mm,
        "exp_time": parameters.exposure_time_s,
        "expt_type": parameters.nexgen_experiment_type,
        "filename": filename_prefix,
        "num_imgs": total_numb_imgs,
        "pump_status": pump_status,
        "pump_exp": parameters.laser_dwell_s,
        "pump_delay": parameters.laser_delay_s,
        "transmission": parameters.transmission,
        "visitpath": os.fspath(meta_h5.parent),
        "wavelength": wavelength_in_a,
        "bit_depth": bit_depth,
        "start_time": start_time,
    }
    SSX_LOGGER.info(f"Sending POST request to {url} with payload:")
    SSX_LOGGER.info(pprint.pformat(payload))
    response = requests.post(url, headers=headers, json=payload)
    response.raise_for_status()
    SSX_LOGGER.info(f"Response: {response.text} (status code: {response.status_code})")



import json
from math import isclose
from pathlib import PosixPath
from unittest.mock import MagicMock, patch

import pytest

from mx_bluesky.common.parameters.constants import GridscanParamConstants
from mx_bluesky.hyperion.external_interaction.agamemnon import (
    PinType,
    SinglePin,
    compare_params,
    get_next_instruction,
    get_pin_type_from_agamemnon_parameters,
    get_withenergy_parameters_from_agamemnon,
    get_withvisit_parameters_from_agamemnon,
    populate_parameters_from_agamemnon,
    update_params_from_agamemnon,
)
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect


@pytest.mark.parametrize(
    "num_wells, well_width, buffer, expected_width",
    [
        (3, 500, 0, 1000),
        (6, 50, 100, 450),
        (2, 800, 50, 900),
    ],
)
def test_given_various_pin_formats_then_pin_width_as_expected(
    num_wells, well_width, buffer, expected_width
):
    pin = PinType(num_wells, well_width, buffer)
    assert pin.full_width == expected_width


def set_up_agamemnon_params(
    loop_type: str | None = None,
    prefix: str | None = None,
    distance: int | None = None,
    wavelength: float | None = None,
):
    return {
        "collection": [{"distance": distance, "wavelength": wavelength}],
        "prefix": prefix,
        "sample": {"loopType": loop_type, "id": 1, "position": 1, "container": 1},
    }


def test_given_no_loop_type_in_parameters_then_single_pin_returned():
    assert (
        get_pin_type_from_agamemnon_parameters(set_up_agamemnon_params()) == SinglePin()
    )


@pytest.mark.parametrize(
    "loop_name, expected_loop",
    [
        ("multipin_6x50+9", PinType(6, 50, 9)),
        ("multipin_6x25.8+8.6", PinType(6, 25.8, 8.6)),
        ("multipin_9x31+90", PinType(9, 31, 90)),
    ],
)
def test_given_multipin_loop_type_in_parameters_then_expected_pin_returned(
    loop_name: str, expected_loop: PinType
):
    assert (
        get_pin_type_from_agamemnon_parameters(set_up_agamemnon_params(loop_name))
        == expected_loop
    )


@pytest.mark.parametrize(
    "loop_name",
    [
        "nonesense",
        "single_pin_78x89+1",
    ],
)
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.LOGGER")
def test_given_completely_unrecognised_loop_type_in_parameters_then_warning_logged_single_pin_returned(
    mock_logger: MagicMock,
    loop_name: str,
):
    assert (
        get_pin_type_from_agamemnon_parameters(set_up_agamemnon_params(loop_name))
        == SinglePin()
    )
    mock_logger.warning.assert_called_once()


@pytest.mark.parametrize(
    "loop_name",
    [
        "multipin_67x56",
        "multipin_90+4",
        "multipin_8",
        "multipin_6x50+",
        "multipin_6x50+98.",
        "multipin_6x50+.1",
        "multipin_6x.50+98",
        "multipin_6x50+98.1.2",
        "multipin_6x50.5.6+98",
        "multipin_6x50+98..1",
        "multipin_6x.50+.98",
        "multipin_6x+98",
    ],
)
def test_given_unrecognised_multipin_in_parameters_then_warning_logged_single_pin_returned(
    loop_name: str,
):
    with pytest.raises(ValueError) as e:
        get_pin_type_from_agamemnon_parameters(set_up_agamemnon_params(loop_name))
    assert "Expected multipin format" in str(e.value)


def configure_mock_agamemnon(mock_requests: MagicMock, loop_type: str | None):
    mock_requests.get.return_value.content = json.dumps(
        {"collect": set_up_agamemnon_params(loop_type, "", 255, 0.9)}
    )


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_when_get_next_instruction_called_then_expected_agamemnon_url_queried(
    mock_requests: MagicMock,
):
    configure_mock_agamemnon(mock_requests, None)
    get_next_instruction("i03")
    mock_requests.get.assert_called_once_with(
        "http://agamemnon.diamond.ac.uk/getnextcollect/i03",
        headers={"Accept": "application/json"},
    )


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_given_agamemnon_returns_an_unexpected_response_then_exception_is_thrown(
    mock_requests: MagicMock,
):
    mock_requests.get.return_value.content = json.dumps({"not_collect": ""})
    with pytest.raises(KeyError) as e:
        get_next_instruction("i03")
    assert "not_collect" in str(e.value)


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_given_agamemnon_returns_multipin_when_get_next_pin_type_from_agamemnon_called_then_multipin_returned(
    mock_requests: MagicMock,
):
    configure_mock_agamemnon(mock_requests, "multipin_6x50+98.1")
    params = get_next_instruction("i03")
    assert get_pin_type_from_agamemnon_parameters(params) == PinType(6, 50, 98.1)


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_given_agamemnon_fails_when_update_parameters_called_then_parameters_unchanged(
    mock_requests: MagicMock, load_centre_collect_params: LoadCentreCollect
):
    mock_requests.get.side_effect = Exception("Bad")
    old_grid_width = load_centre_collect_params.robot_load_then_centre.grid_width_um
    params = update_params_from_agamemnon(load_centre_collect_params)
    assert params.robot_load_then_centre.grid_width_um == old_grid_width


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.compare_params")
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_given_agamemnon_gives_single_pin_when_update_parameters_called_then_parameters_changed_to_single_pin(
    mock_requests: MagicMock,
    mock_compare_params: MagicMock,
    load_centre_collect_params: LoadCentreCollect,
):
    configure_mock_agamemnon(mock_requests, None)
    load_centre_collect_params.robot_load_then_centre.grid_width_um = 0
    load_centre_collect_params.select_centres.n = 0
    params = update_params_from_agamemnon(load_centre_collect_params)
    assert (
        params.robot_load_then_centre.grid_width_um == GridscanParamConstants.WIDTH_UM
    )
    assert params.select_centres.n == 1
    assert params.multi_rotation_scan.snapshot_omegas_deg


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.compare_params")
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_given_agamemnon_gives_multi_pin_when_update_parameters_called_then_parameters_changed_to_multi_pin(
    mock_requests: MagicMock,
    mock_compare_params: MagicMock,
    load_centre_collect_params: LoadCentreCollect,
):
    configure_mock_agamemnon(mock_requests, "multipin_6x50+10")
    params = update_params_from_agamemnon(load_centre_collect_params)
    assert params.robot_load_then_centre.grid_width_um == 270
    assert params.select_centres.n == 6
    assert params.robot_load_then_centre.tip_offset_um == 135
    assert not params.multi_rotation_scan.snapshot_omegas_deg


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_given_set_of_parameters_then_correct_agamemnon_url_is_deduced(
    mock_requests: MagicMock, load_centre_collect_params: LoadCentreCollect
):
    update_params_from_agamemnon(load_centre_collect_params)
    mock_requests.get.assert_called_once_with(
        "http://agamemnon.diamond.ac.uk/getnextcollect/i03",
        headers={"Accept": "application/json"},
    )


@pytest.mark.parametrize(
    "prefix, expected_visit",
    [
        ["/dls/i03/data/2025/mx23694-130/foo/bar", "mx23694-130"],
        ["/dls/not-i03/data/2021/mx84743-230", "mx84743-230"],
    ],
)
def test_given_valid_prefix_then_correct_visit_is_set(prefix: str, expected_visit: str):
    visit, _ = get_withvisit_parameters_from_agamemnon(
        set_up_agamemnon_params(None, prefix, None)
    )
    assert visit == expected_visit


@pytest.mark.parametrize(
    "prefix",
    [
        "/not-dls/i03/data/2025/mx23694-130/foo/bar",
        "/dls/i03/not-data/2025/mx23694-130/foo/bar",
        "/foo/bar/i03/data/2025/mx23694-130",
    ],
)
def test_given_invalid_prefix_then_exception_raised(prefix: str):
    with pytest.raises(ValueError) as e:
        get_withvisit_parameters_from_agamemnon(
            set_up_agamemnon_params(None, prefix, None)
        )

    assert "MX-General root structure" in str(e.value)


def test_no_prefix_raises_exception():
    with pytest.raises(KeyError) as e:
        get_withvisit_parameters_from_agamemnon({"not_collect": ""})

    assert "Unexpected json from agamemnon" in str(e.value)


@pytest.mark.parametrize(
    "mock_error, mock_log",
    [
        (ValueError(), "Failed to compare parameters: "),
        (Exception(), "Unexpected error occurred. Failed to compare parameters: "),
    ],
)
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.LOGGER")
@patch(
    "mx_bluesky.hyperion.external_interaction.agamemnon.populate_parameters_from_agamemnon"
)
def test_if_failed_to_populate_parameters_from_hyperion_exception_is_logged(
    mock_populate_params,
    mock_logger,
    mock_requests,
    mock_error,
    mock_log,
    load_centre_collect_params: LoadCentreCollect,
):
    configure_mock_agamemnon(mock_requests, None)
    mock_populate_params.side_effect = mock_error
    compare_params(
        load_centre_collect_params,
    )
    mock_logger.warning.assert_called_with(mock_log)


@pytest.mark.parametrize(
    "agamemnon_params",
    ["example_collect.json", "example_collect_multipin.json"],
)
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.LOGGER")
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
@patch("mx_bluesky.common.parameters.components.os", new=MagicMock())
def test_populate_parameters_from_agamemnon_causes_no_warning_when_compared_to_gda_params(
    mock_requests: MagicMock,
    mock_logger: MagicMock,
    agamemnon_params: str,
    load_centre_collect_params: LoadCentreCollect,
):
    with open(f"tests/test_data/agamemnon/{agamemnon_params}") as json_file:
        example_json = json_file.read()
        mock_requests.get.return_value.content = example_json

    compare_params(load_centre_collect_params)
    mock_logger.warning.assert_not_called()


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
@patch("mx_bluesky.common.parameters.components.os", new=MagicMock())
def test_populate_parameters_from_agamemnon_contains_expected_data(
    mock_requests: MagicMock,
):
    with open("tests/test_data/agamemnon/example_collect.json") as json_file:
        example_json = json_file.read()
        mock_requests.get.return_value.content = example_json

    agamemnon_params = get_next_instruction("i03")
    hyperion_params = populate_parameters_from_agamemnon(agamemnon_params)
    assert hyperion_params.visit == "cm00000-0"
    assert isclose(hyperion_params.detector_distance_mm, 180.8)  # type: ignore
    assert hyperion_params.sample_id == 12345
    assert hyperion_params.sample_puck == 40
    assert hyperion_params.sample_pin == 3
    assert str(hyperion_params.parameter_model_version) == "5.3.0"
    assert hyperion_params.select_centres.n == 1


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
@patch("mx_bluesky.common.parameters.components.os", new=MagicMock())
def test_populate_parameters_from_agamemnon_contains_expected_robot_load_then_centre_data(
    mock_requests: MagicMock,
):
    with open("tests/test_data/agamemnon/example_collect.json") as json_file:
        example_json = json_file.read()
        mock_requests.get.return_value.content = example_json

    agamemnon_params = get_next_instruction("i03")
    hyperion_params = populate_parameters_from_agamemnon(agamemnon_params)
    robot_load_params = hyperion_params.robot_load_then_centre
    assert robot_load_params.visit == "cm00000-0"
    assert isclose(robot_load_params.detector_distance_mm, 180.8)  # type: ignore
    assert robot_load_params.sample_id == 12345
    assert robot_load_params.sample_puck == 40
    assert robot_load_params.sample_pin == 3
    assert robot_load_params.demand_energy_ev == 12700.045934258673
    assert str(robot_load_params.parameter_model_version) == "5.3.0"
    assert (
        robot_load_params.storage_directory
        == "/dls/tmp/data/year/cm00000-0/auto/test/xraycentring"
    )
    assert robot_load_params.file_name == "test_xtal"
    assert robot_load_params.snapshot_directory == PosixPath(
        "/dls/tmp/data/year/cm00000-0/auto/test/snapshots"
    )


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
@patch("mx_bluesky.common.parameters.components.os", new=MagicMock())
def test_populate_multipin_parameters_from_agamemnon(
    mock_requests: MagicMock,
):
    with open("tests/test_data/agamemnon/example_collect_multipin.json") as json_file:
        example_json = json_file.read()
        mock_requests.get.return_value.content = example_json

    agamemnon_params = get_next_instruction("i03")
    hyperion_params = populate_parameters_from_agamemnon(agamemnon_params)
    assert hyperion_params.select_centres.n == 6


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
@patch("mx_bluesky.common.parameters.components.os", new=MagicMock())
def test_get_withenergy_parameters_from_agamemnon(mock_requests: MagicMock):
    with open("tests/test_data/agamemnon/example_collect.json") as json_file:
        example_json = json_file.read()
        mock_requests.get.return_value.content = example_json

    agamemnon_params = get_next_instruction("i03")
    demand_energy_ev = get_withenergy_parameters_from_agamemnon(agamemnon_params)
    assert demand_energy_ev["demand_energy_ev"] == 12700.045934258673


def test_get_withenergy_parameters_from_agamemnon_when_no_wavelength():
    agamemnon_params = {}
    demand_energy_ev = get_withenergy_parameters_from_agamemnon(agamemnon_params)
    assert demand_energy_ev["demand_energy_ev"] is None



#!/usr/bin/env python3
# This script requires python >= 3.10 but we can't warn about this because
# it will fail at compile time before we can execute a version check
import subprocess
import sys
from pathlib import Path
from shutil import copyfile
from tempfile import TemporaryDirectory

import h5py


def main() -> int:
    filenames = []
    for option in sys.argv[1:]:
        match option:
            case "--help" | "-h":
                command = sys.argv[0]
                print(
                    f"{command} [metafile]"
                    f"\n\tStrip the /flatfield and /mask from an HDF5 meta file to make it much smaller"
                    f"\n\tTHIS WILL MODIFY THE FILE"
                    f"\n{command} -h | --help"
                    f"\n\tThis help"
                )
                return 0
            case arg:
                filenames.append(arg)

    if len(filenames) < 1:
        sys.stderr.write("Input and/or output file name not supplied\n")
        return 1

    inputfile = filenames[0]
    inputpath = Path(inputfile)

    with TemporaryDirectory() as tempdir:
        tmpfile = f"{tempdir}/{inputpath.name}"
        copyfile(inputfile, tmpfile)
        with h5py.File(tmpfile, "r+") as metafile:
            del metafile["flatfield"]
            metafile.create_dataset("flatfield", (4362, 4148), "<f4")
            del metafile["mask"]
            metafile.create_dataset("mask", (4362, 4148), "<i4")

        h5repack(tempdir, tmpfile, inputfile)

    return 0


def h5repack(tempdir, file, inputfile):
    tempfile = f"{tempdir}/repacked.h5"
    run_command(["h5repack", file, tempfile])
    run_command(["gzip", "--best", tempfile])
    copyfile(f"{tempfile}.gz", f"{inputfile}.gz")


def run_command(args):
    completion = subprocess.run(args, capture_output=True, text=True)
    if completion.returncode:
        sys.stderr.write(f"{args[0]} failed with return code {completion.returncode}")
        sys.stderr.write(completion.stderr)
        sys.exit(1)
    return completion


if __name__ == "__main__":
    sys.exit(main())



from pydantic import Field

from mx_bluesky.common.parameters.components import (
    MxBlueskyParameters,
    WithOptionalEnergyChange,
    WithSample,
    WithSnapshot,
    WithVisit,
)
from mx_bluesky.common.parameters.constants import (
    HardwareConstants,
)
from mx_bluesky.hyperion.parameters.components import WithHyperionUDCFeatures
from mx_bluesky.hyperion.parameters.gridscan import (
    GridCommonWithHyperionDetectorParams,
    PinTipCentreThenXrayCentre,
)


class RobotLoadAndEnergyChange(
    MxBlueskyParameters, WithSample, WithSnapshot, WithOptionalEnergyChange, WithVisit
):
    thawing_time: float = Field(default=HardwareConstants.THAWING_TIME)


class RobotLoadThenCentre(
    GridCommonWithHyperionDetectorParams, WithHyperionUDCFeatures
):
    thawing_time: float = Field(default=HardwareConstants.THAWING_TIME)

    @property
    def robot_load_params(self) -> RobotLoadAndEnergyChange:
        my_params = self.model_dump()
        return RobotLoadAndEnergyChange(**my_params)

    @property
    def pin_centre_then_xray_centre_params(self) -> PinTipCentreThenXrayCentre:
        my_params = self.model_dump()
        del my_params["thawing_time"]
        return PinTipCentreThenXrayCentre(**my_params)



import os
import re
from decimal import Decimal
from functools import partial
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from aiohttp import ClientResponse
from dodal.beamlines import i03
from dodal.devices.oav.oav_parameters import OAVConfig
from ophyd_async.core import AsyncStatus
from ophyd_async.testing import set_mock_value
from PIL import Image

from mx_bluesky.hyperion.parameters.constants import CONST

# Map all the case-sensitive column names from their normalised versions
DATA_COLLECTION_COLUMN_MAP = {
    s.lower(): s
    for s in [
        "dataCollectionId",
        "BLSAMPLEID",
        "SESSIONID",
        "experimenttype",
        "dataCollectionNumber",
        "startTime",
        "endTime",
        "runStatus",
        "axisStart",
        "axisEnd",
        "axisRange",
        "overlap",
        "numberOfImages",
        "startImageNumber",
        "numberOfPasses",
        "exposureTime",
        "imageDirectory",
        "imagePrefix",
        "imageSuffix",
        "imageContainerSubPath",
        "fileTemplate",
        "wavelength",
        "resolution",
        "detectorDistance",
        "xBeam",
        "yBeam",
        "comments",
        "printableForReport",
        "CRYSTALCLASS",
        "slitGapVertical",
        "slitGapHorizontal",
        "transmission",
        "synchrotronMode",
        "xtalSnapshotFullPath1",
        "xtalSnapshotFullPath2",
        "xtalSnapshotFullPath3",
        "xtalSnapshotFullPath4",
        "rotationAxis",
        "phiStart",
        "kappaStart",
        "omegaStart",
        "chiStart",
        "resolutionAtCorner",
        "detector2Theta",
        "DETECTORMODE",
        "undulatorGap1",
        "undulatorGap2",
        "undulatorGap3",
        "beamSizeAtSampleX",
        "beamSizeAtSampleY",
        "centeringMethod",
        "averageTemperature",
        "ACTUALSAMPLEBARCODE",
        "ACTUALSAMPLESLOTINCONTAINER",
        "ACTUALCONTAINERBARCODE",
        "ACTUALCONTAINERSLOTINSC",
        "actualCenteringPosition",
        "beamShape",
        "dataCollectionGroupId",
        "POSITIONID",
        "detectorId",
        "FOCALSPOTSIZEATSAMPLEX",
        "POLARISATION",
        "FOCALSPOTSIZEATSAMPLEY",
        "APERTUREID",
        "screeningOrigId",
        "flux",
        "strategySubWedgeOrigId",
        "blSubSampleId",
        "processedDataFile",
        "datFullPath",
        "magnification",
        "totalAbsorbedDose",
        "binning",
        "particleDiameter",
        "boxSize",
        "minResolution",
        "minDefocus",
        "maxDefocus",
        "defocusStepSize",
        "amountAstigmatism",
        "extractSize",
        "bgRadius",
        "voltage",
        "objAperture",
        "c1aperture",
        "c2aperture",
        "c3aperture",
        "c1lens",
        "c2lens",
        "c3lens",
        "startPositionId",
        "endPositionId",
        "flux",
        "bestWilsonPlotPath",
        "totalExposedDose",
        "nominalMagnification",
        "nominalDefocus",
        "imageSizeX",
        "imageSizeY",
        "pixelSizeOnImage",
        "phasePlate",
        "dataCollectionPlanId",
    ]
}


@pytest.fixture(autouse=True)
def use_dev_ispyb_unless_overridden_by_environment():
    ispyb_config_path = os.environ.get(
        "ISPYB_CONFIG_PATH", CONST.SIM.DEV_ISPYB_DATABASE_CFG
    )
    with patch.dict(os.environ, {"ISPYB_CONFIG_PATH": ispyb_config_path}):
        yield ispyb_config_path


@pytest.fixture
def undulator_for_system_test(undulator):
    set_mock_value(undulator.current_gap, 1.11)
    return undulator


@pytest.fixture
def oav_for_system_test(test_config_files):
    parameters = OAVConfig(
        test_config_files["zoom_params_file"], test_config_files["display_config"]
    )
    oav = i03.oav(connect_immediately=True, mock=True, params=parameters)
    set_mock_value(oav.cam.array_size_x, 1024)
    set_mock_value(oav.cam.array_size_y, 768)

    # Grid snapshots
    set_mock_value(oav.grid_snapshot.x_size, 1024)
    set_mock_value(oav.grid_snapshot.y_size, 768)
    set_mock_value(oav.grid_snapshot.top_left_x, 50)
    set_mock_value(oav.grid_snapshot.top_left_y, 100)
    size_in_pixels = 0.1 * 1000 / 1.25
    set_mock_value(oav.grid_snapshot.box_width, size_in_pixels)

    # Rotation snapshots
    @AsyncStatus.wrap
    async def trigger_with_test_image(self):
        with Image.open(
            "tests/test_data/test_images/generate_snapshot_input.png"
        ) as image:
            await self.post_processing(image)

    oav.snapshot.trigger = MagicMock(
        side_effect=partial(trigger_with_test_image, oav.snapshot)
    )
    oav.grid_snapshot.trigger = MagicMock(
        side_effect=partial(trigger_with_test_image, oav.grid_snapshot)
    )

    empty_response = AsyncMock(spec=ClientResponse)
    empty_response.read.return_value = b""

    with (
        patch(
            "dodal.devices.areadetector.plugins.MJPG.ClientSession.get", autospec=True
        ) as mock_get,
    ):
        mock_get.return_value.__aenter__.return_value = empty_response
        set_mock_value(oav.zoom_controller.level, "1.0")
        zoom_levels_list = ["1.0x", "3.0x", "5.0x", "7.5x", "10.0x", "15.0x"]
        oav.zoom_controller._get_allowed_zoom_levels = AsyncMock(
            return_value=zoom_levels_list
        )
        yield oav


def compare_actual_and_expected(
    id, expected_values, fetch_datacollection_attribute, column_map: dict | None = None
):
    results = "\n"
    for k, v in expected_values.items():
        actual = fetch_datacollection_attribute(
            id, column_map[k.lower()] if column_map else k
        )
        if isinstance(actual, Decimal):
            actual = float(actual)
        if isinstance(v, float):
            actual_v = actual == pytest.approx(v)
        elif isinstance(v, str) and v.startswith("regex:"):
            actual_v = re.match(v.removeprefix("regex:"), str(actual))  # type: ignore
        else:
            actual_v = actual == v
        if not actual_v:
            results += f"expected {k} {v} == {actual}\n"
    assert results == "\n", results


def compare_comment(
    fetch_datacollection_attribute, data_collection_id, expected_comment
):
    actual_comment = fetch_datacollection_attribute(
        data_collection_id, DATA_COLLECTION_COLUMN_MAP["comments"]
    )
    match = re.search(" Zocalo processing took", actual_comment)
    truncated_comment = actual_comment[: match.start()] if match else actual_comment
    assert truncated_comment == expected_comment



from __future__ import annotations

from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGroupInfo,
    DataCollectionInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    EIGER_FILE_SUFFIX,
    I03_EIGER_DETECTOR,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_utils import (
    get_current_time_string,
)
from mx_bluesky.common.parameters.components import DiffractionExperimentWithSample


def populate_data_collection_group(params: DiffractionExperimentWithSample):
    dcg_info = DataCollectionGroupInfo(
        visit_string=params.visit,
        experiment_type=params.ispyb_experiment_type.value,
        sample_id=params.sample_id,
    )
    return dcg_info


def populate_remaining_data_collection_info(
    comment,
    data_collection_group_id,
    data_collection_info: DataCollectionInfo,
    params: DiffractionExperimentWithSample,
):
    data_collection_info.visit_string = params.visit
    data_collection_info.parent_id = data_collection_group_id
    data_collection_info.sample_id = params.sample_id
    data_collection_info.detector_id = I03_EIGER_DETECTOR
    data_collection_info.comments = comment
    data_collection_info.detector_distance = params.detector_params.detector_distance
    data_collection_info.exp_time = params.detector_params.exposure_time_s
    data_collection_info.imgdir = params.detector_params.directory
    data_collection_info.imgprefix = params.detector_params.prefix
    data_collection_info.imgsuffix = EIGER_FILE_SUFFIX
    # Both overlap and n_passes included for backwards compatibility,
    # planned to be removed later
    data_collection_info.n_passes = 1
    data_collection_info.overlap = 0
    data_collection_info.start_image_number = 1
    beam_position = params.detector_params.get_beam_position_mm(
        params.detector_params.detector_distance
    )
    data_collection_info.xbeam = beam_position[0]
    data_collection_info.ybeam = beam_position[1]
    data_collection_info.start_time = get_current_time_string()
    # temporary file template until nxs filewriting is integrated and we can use
    # that file name
    data_collection_info.file_template = f"{params.detector_params.prefix}_{data_collection_info.data_collection_number}_master.h5"
    return data_collection_info


def get_proposal_and_session_from_visit_string(visit_string: str) -> tuple[str, int]:
    visit_parts = visit_string.split("-")
    assert len(visit_parts) == 2, f"Unexpected visit string {visit_string}"
    return visit_parts[0], int(visit_parts[1])



from __future__ import annotations

from pathlib import Path

import pydantic
from blueapi.core import BlueskyContext
from bluesky import plan_stubs as bps
from bluesky import preprocessors as bpp
from bluesky.preprocessors import subs_decorator
from bluesky.utils import MsgGenerator
from dodal.devices.aperturescatterguard import ApertureScatterguard
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight, BacklightPosition
from dodal.devices.dcm import DCM
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import PandAFastGridScan, ZebraFastGridScan
from dodal.devices.flux import Flux
from dodal.devices.i03.beamstop import Beamstop
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.undulator import Undulator
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from dodal.devices.zocalo import ZocaloResults
from ophyd_async.fastcs.panda import HDFPanda

from mx_bluesky.common.external_interaction.callbacks.common.grid_detection_callback import (
    GridDetectionCallback,
    GridParamUpdate,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    ispyb_activation_wrapper,
)
from mx_bluesky.common.parameters.constants import OavConstants
from mx_bluesky.common.utils.context import device_composite_from_context
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.xrc_result import XRayCentreEventHandler
from mx_bluesky.hyperion.device_setup_plans.manipulate_sample import (
    move_aperture_if_required,
)
from mx_bluesky.hyperion.device_setup_plans.utils import (
    start_preparing_data_collection_then_do_plan,
)
from mx_bluesky.hyperion.experiment_plans.change_aperture_then_move_plan import (
    change_aperture_then_move_to_xtal,
)
from mx_bluesky.hyperion.experiment_plans.flyscan_xray_centre_plan import (
    flyscan_xray_centre_no_move,
)
from mx_bluesky.hyperion.experiment_plans.oav_grid_detection_plan import (
    OavGridDetectionComposite,
    grid_detection_plan,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionFlyScanXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import (
    GridScanWithEdgeDetect,
    HyperionSpecifiedThreeDGridScan,
)


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class GridDetectThenXRayCentreComposite:
    """All devices which are directly or indirectly required by this plan"""

    aperture_scatterguard: ApertureScatterguard
    attenuator: BinaryFilterAttenuator
    backlight: Backlight
    beamstop: Beamstop
    dcm: DCM
    detector_motion: DetectorMotion
    eiger: EigerDetector
    zebra_fast_grid_scan: ZebraFastGridScan
    flux: Flux
    oav: OAV
    pin_tip_detection: PinTipDetection
    smargon: Smargon
    synchrotron: Synchrotron
    s4_slit_gaps: S4SlitGaps
    undulator: Undulator
    xbpm_feedback: XBPMFeedback
    zebra: Zebra
    zocalo: ZocaloResults
    panda: HDFPanda
    panda_fast_grid_scan: PandAFastGridScan
    robot: BartRobot
    sample_shutter: ZebraShutter


def create_devices(context: BlueskyContext) -> GridDetectThenXRayCentreComposite:
    return device_composite_from_context(context, GridDetectThenXRayCentreComposite)


def create_parameters_for_flyscan_xray_centre(
    grid_scan_with_edge_params: GridScanWithEdgeDetect,
    grid_parameters: GridParamUpdate,
) -> HyperionSpecifiedThreeDGridScan:
    params_json = grid_scan_with_edge_params.model_dump()
    params_json.update(grid_parameters)
    flyscan_xray_centre_parameters = HyperionSpecifiedThreeDGridScan(**params_json)
    LOGGER.info(f"Parameters for FGS: {flyscan_xray_centre_parameters}")
    return flyscan_xray_centre_parameters


def detect_grid_and_do_gridscan(
    composite: GridDetectThenXRayCentreComposite,
    parameters: GridScanWithEdgeDetect,
    oav_params: OAVParameters,
):
    snapshot_template = f"{parameters.detector_params.prefix}_{parameters.detector_params.run_number}_{{angle}}"

    grid_params_callback = GridDetectionCallback()

    @bpp.subs_decorator([grid_params_callback])
    def run_grid_detection_plan(
        oav_params,
        snapshot_template,
        snapshot_dir: Path,
    ):
        grid_detect_composite = OavGridDetectionComposite(
            backlight=composite.backlight,
            oav=composite.oav,
            smargon=composite.smargon,
            pin_tip_detection=composite.pin_tip_detection,
        )

        yield from grid_detection_plan(
            grid_detect_composite,
            oav_params,
            snapshot_template,
            str(snapshot_dir),
            parameters.grid_width_um,
            parameters.box_size_um,
        )

    if parameters.selected_aperture:
        # Start moving the aperture/scatterguard into position without moving it in
        yield from bps.prepare(
            composite.aperture_scatterguard,
            parameters.selected_aperture,
            group=CONST.WAIT.GRID_READY_FOR_DC,
        )

    yield from run_grid_detection_plan(
        oav_params,
        snapshot_template,
        parameters.snapshot_directory,
    )

    yield from bps.abs_set(
        composite.backlight, BacklightPosition.OUT, group=CONST.WAIT.GRID_READY_FOR_DC
    )

    yield from move_aperture_if_required(
        composite.aperture_scatterguard,
        parameters.selected_aperture,
        group=CONST.WAIT.GRID_READY_FOR_DC,
    )

    yield from flyscan_xray_centre_no_move(
        HyperionFlyScanXRayCentreComposite(
            aperture_scatterguard=composite.aperture_scatterguard,
            attenuator=composite.attenuator,
            backlight=composite.backlight,
            eiger=composite.eiger,
            panda_fast_grid_scan=composite.panda_fast_grid_scan,
            flux=composite.flux,
            s4_slit_gaps=composite.s4_slit_gaps,
            smargon=composite.smargon,
            undulator=composite.undulator,
            synchrotron=composite.synchrotron,
            xbpm_feedback=composite.xbpm_feedback,
            zebra=composite.zebra,
            zocalo=composite.zocalo,
            panda=composite.panda,
            zebra_fast_grid_scan=composite.zebra_fast_grid_scan,
            dcm=composite.dcm,
            robot=composite.robot,
            sample_shutter=composite.sample_shutter,
        ),
        create_parameters_for_flyscan_xray_centre(
            parameters, grid_params_callback.get_grid_parameters()
        ),
    )


def grid_detect_then_xray_centre(
    composite: GridDetectThenXRayCentreComposite,
    parameters: GridScanWithEdgeDetect,
    oav_config: str = OavConstants.OAV_CONFIG_JSON,
) -> MsgGenerator:
    """
    A plan which combines the collection of snapshots from the OAV and the determination
    of the grid dimensions to use for the following grid scan.
    """

    eiger: EigerDetector = composite.eiger

    eiger.set_detector_parameters(parameters.detector_params)

    oav_params = OAVParameters("xrayCentring", oav_config)

    flyscan_event_handler = XRayCentreEventHandler()

    @subs_decorator(flyscan_event_handler)
    def plan_to_perform():
        yield from ispyb_activation_wrapper(
            detect_grid_and_do_gridscan(
                composite,
                parameters,
                oav_params,
            ),
            parameters,
        )

    yield from start_preparing_data_collection_then_do_plan(
        composite.beamstop,
        eiger,
        composite.detector_motion,
        parameters.detector_params.detector_distance,
        plan_to_perform(),
        group=CONST.WAIT.GRID_READY_FOR_DC,
    )

    assert flyscan_event_handler.xray_centre_results, (
        "Flyscan result event not received or no crystal found and exception not raised"
    )

    yield from change_aperture_then_move_to_xtal(
        flyscan_event_handler.xray_centre_results[0],
        composite.smargon,
        composite.aperture_scatterguard,
    )



from collections.abc import Iterator
from os import environ, getenv
from unittest.mock import patch

import pytest

environ["HYPERION_TEST_MODE"] = "true"

print("Adjusting S03 EPICS environment ...")
s03_epics_server_port = getenv("S03_EPICS_CA_SERVER_PORT")
s03_epics_repeater_port = getenv("S03_EPICS_CA_REPEATER_PORT")
if s03_epics_server_port:
    environ["EPICS_CA_SERVER_PORT"] = s03_epics_server_port
    print(f"[EPICS_CA_SERVER_PORT] = {s03_epics_server_port}")
if s03_epics_repeater_port:
    environ["EPICS_CA_REPEATER_PORT"] = s03_epics_repeater_port
    print(f"[EPICS_CA_REPEATER_PORT] = {s03_epics_repeater_port}")


def pytest_addoption(parser):
    parser.addoption(
        "--logging",
        action="store_true",
        default=False,
        help="Log during all tests (not just those that are testing logging logic)",
    )


@pytest.fixture(scope="session", autouse=True)
def default_session_fixture() -> Iterator[None]:
    print("Patching bluesky 0MQ Publisher in __main__ for the whole session")
    with patch("mx_bluesky.hyperion.__main__.Publisher"):
        yield



import re

from dodal.devices.oav.snapshots.snapshot_image_processing import (
    compute_beam_centre_pixel_xy_for_mm_position,
    draw_crosshair,
)
from event_model import Event, EventDescriptor, RunStart
from PIL import Image

from mx_bluesky.common.external_interaction.callbacks.common.plan_reactive_callback import (
    PlanReactiveCallback,
)
from mx_bluesky.common.parameters.constants import DocDescriptorNames
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER as CALLBACK_LOGGER


class BeamDrawingCallback(PlanReactiveCallback):
    """
    Callback that monitors for OAV_ROTATION_SNAPSHOT_TRIGGERED events and
    draws a crosshair at the beam centre, saving the snapshot to a file.
    The callback assumes an OAV device "oav"
    Examples:
        Take a snapshot at the current location
    >>> from bluesky.run_engine import RunEngine
    >>> import bluesky.preprocessors as bpp
    >>> import bluesky.plan_stubs as bps
    >>> from dodal.devices.oav.oav_detector import OAV
    >>> from mx_bluesky.common.parameters.components import WithSnapshot
    >>> def take_snapshot(params: WithSnapshot, oav: OAV, run_engine: RunEngine):
    ...     run_engine.subscribe(BeamDrawingCallback())
    ...     @bpp.run_decorator(md={
    ...     "activate_callbacks": ["BeamDrawingCallback"],
    ...         "with_snapshot": params.model_dump_json(),
    ...     })
    ...     def inner_plan():
    ...         yield from bps.abs_set(oav.snapshot.directory, "/path/to/snapshot_folder", wait=True)
    ...         yield from bps.abs_set(oav.snapshot.filename, "my_snapshot_prefix", wait=True)
    ...         yield from bps.trigger(oav.snapshot, wait=True)
    ...         yield from bps.create(DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED)
    ...         yield from bps.read(oav)
    ...         yield from bps.save()
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, log=CALLBACK_LOGGER, **kwargs)
        self._snapshot_files: list[str] = []
        self._microns_per_pixel: tuple[float, float]
        self._beam_centre: tuple[int, int]
        self._rotation_snapshot_descriptor: str = ""

    def activity_gated_start(self, doc: RunStart):
        if self.activity_uid == doc.get("uid"):
            with_snapshot_json = doc.get("with_snapshot")  # type: ignore
            assert with_snapshot_json, (
                "run start event did not have expected snapshot json"
            )
        return doc

    def activity_gated_descriptor(self, doc: EventDescriptor) -> EventDescriptor | None:
        if doc.get("name") == DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED:
            self._rotation_snapshot_descriptor = doc["uid"]
        return doc

    def activity_gated_event(self, doc: Event) -> Event:
        if doc["descriptor"] == self._rotation_snapshot_descriptor:
            self._handle_rotation_snapshot(doc)
        return doc

    def _extract_base_snapshot_params(self, doc: Event):
        data = doc["data"]
        self._snapshot_files.append(data["oav-snapshot-last_saved_path"])
        self._microns_per_pixel = (
            data["oav-microns_per_pixel_x"],
            data["oav-microns_per_pixel_y"],
        )
        self._beam_centre = (data["oav-beam_centre_i"], data["oav-beam_centre_j"])

    def _handle_rotation_snapshot(self, doc: Event):
        self._extract_base_snapshot_params(doc)
        data = doc["data"]
        snapshot_path = data["oav-snapshot-last_saved_path"]
        match = re.match("(.*)\\.png", snapshot_path)
        assert match, f"Snapshot {snapshot_path} was not a .png file"
        snapshot_base = match.groups()[0]
        output_snapshot_path = f"{snapshot_base}_with_beam_centre.png"
        self._generate_snapshot_at(snapshot_path, output_snapshot_path, 0, 0)
        data["oav-snapshot-last_saved_path"] = output_snapshot_path
        return doc

    def _generate_snapshot_at(
        self, input_snapshot_path: str, output_snapshot_path: str, x_mm: int, y_mm: int
    ):
        """
        Save a snapshot to the specified path, with an annotated crosshair at the specified
        position
        Args:
            input_snapshot_path: The non-annotated image path.
            output_snapshot_path:  The path to the image that will be annotated.
            x_mm: Relative x location of the sample to the original image (mm)
            y_mm: Relative y location of the sample to the original image (mm)
        """
        image = Image.open(input_snapshot_path)
        x_px, y_px = compute_beam_centre_pixel_xy_for_mm_position(
            (x_mm, y_mm), self._beam_centre, self._microns_per_pixel
        )
        draw_crosshair(image, x_px, y_px)
        image.save(output_snapshot_path, format="png")



from __future__ import annotations

import dataclasses
from collections.abc import Callable, Sequence
from functools import partial
from pathlib import Path
from typing import Protocol

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import numpy as np
from blueapi.core import BlueskyContext
from bluesky.utils import MsgGenerator
from dodal.devices.fast_grid_scan import (
    FastGridScanCommon,
)
from dodal.devices.fast_grid_scan import (
    set_fast_grid_scan_params as set_flyscan_params,
)
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zocalo.zocalo_results import (
    ZOCALO_READING_PLAN_NAME,
    ZOCALO_STAGE_GROUP,
    XrcResult,
    get_full_processing_results,
)
from dodal.plans.preprocessors.verify_undulator_gap import (
    verify_undulator_gap_before_run_decorator,
)

from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    ispyb_activation_wrapper,
)
from mx_bluesky.common.parameters.constants import HardwareConstants
from mx_bluesky.common.plans.do_fgs import kickoff_and_complete_gridscan
from mx_bluesky.common.plans.read_hardware import (
    standard_read_hardware_during_collection,
    standard_read_hardware_pre_collection,
)
from mx_bluesky.common.preprocessors.preprocessors import (
    transmission_and_xbpm_feedback_for_collection_decorator,
)
from mx_bluesky.common.utils.context import device_composite_from_context
from mx_bluesky.common.utils.exceptions import (
    CrystalNotFoundException,
    SampleException,
)
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.utils.tracing import TRACER
from mx_bluesky.common.xrc_result import XRayCentreEventHandler, XRayCentreResult
from mx_bluesky.hyperion.device_setup_plans.setup_panda import (
    disarm_panda_for_gridscan,
    set_panda_directory,
    setup_panda_for_flyscan,
)
from mx_bluesky.hyperion.device_setup_plans.setup_zebra import (
    setup_zebra_for_gridscan,
    setup_zebra_for_panda_flyscan,
    tidy_up_zebra_after_gridscan,
)
from mx_bluesky.hyperion.experiment_plans.change_aperture_then_move_plan import (
    change_aperture_then_move_to_xtal,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionFlyScanXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan

ZOCALO_MIN_TOTAL_COUNT_THRESHOLD = 3


class SmargonSpeedException(Exception):
    pass


def create_devices(context: BlueskyContext) -> HyperionFlyScanXRayCentreComposite:
    """Creates the devices required for the plan and connect to them"""
    return device_composite_from_context(context, HyperionFlyScanXRayCentreComposite)


def flyscan_xray_centre_no_move(
    composite: HyperionFlyScanXRayCentreComposite,
    parameters: HyperionSpecifiedThreeDGridScan,
) -> MsgGenerator:
    """Perform a flyscan and determine the centres of interest"""

    composite.eiger.set_detector_parameters(parameters.detector_params)
    composite.zocalo.zocalo_environment = CONST.ZOCALO_ENV

    parameters.features.update_self_from_server()
    composite.zocalo.use_cpu_and_gpu = parameters.features.compare_cpu_and_gpu_zocalo
    composite.zocalo.use_gpu = parameters.features.use_gpu_results

    feature_controlled = _get_feature_controlled(composite, parameters)

    @bpp.set_run_key_decorator(CONST.PLAN.GRIDSCAN_OUTER)
    @bpp.run_decorator(  # attach experiment metadata to the start document
        md={
            "subplan_name": CONST.PLAN.GRIDSCAN_OUTER,
            "mx_bluesky_parameters": parameters.model_dump_json(),
            "activate_callbacks": [
                "GridscanNexusFileCallback",
            ],
        }
    )
    @bpp.finalize_decorator(lambda: feature_controlled.tidy_plan(composite))
    def run_gridscan_and_fetch_and_tidy(
        fgs_composite: HyperionFlyScanXRayCentreComposite,
        params: HyperionSpecifiedThreeDGridScan,
        feature_controlled: _FeatureControlled,
    ) -> MsgGenerator:
        yield from run_gridscan_and_fetch_results(
            fgs_composite, params, feature_controlled
        )

    yield from run_gridscan_and_fetch_and_tidy(
        composite, parameters, feature_controlled
    )


def flyscan_xray_centre(
    composite: HyperionFlyScanXRayCentreComposite,
    parameters: HyperionSpecifiedThreeDGridScan,
) -> MsgGenerator:
    """Create the plan to run the grid scan based on provided parameters.

    The ispyb handler should be added to the whole gridscan as we want to capture errors
    at any point in it.

    Args:
        parameters (HyperionSpecifiedThreeDGridScan): The parameters to run the scan.

    Returns:
        Generator: The plan for the gridscan
    """
    xrc_event_handler = XRayCentreEventHandler()

    @transmission_and_xbpm_feedback_for_collection_decorator(
        composite,
        parameters.transmission_frac,
    )
    @verify_undulator_gap_before_run_decorator(composite)
    @bpp.subs_decorator(xrc_event_handler)
    def flyscan_and_fetch_results() -> MsgGenerator:
        yield from ispyb_activation_wrapper(
            flyscan_xray_centre_no_move(composite, parameters), parameters
        )

    yield from flyscan_and_fetch_results()

    xray_centre_results = xrc_event_handler.xray_centre_results
    assert xray_centre_results, (
        "Flyscan result event not received or no crystal found and exception not raised"
    )
    yield from change_aperture_then_move_to_xtal(
        xray_centre_results[0],
        composite.smargon,
        composite.aperture_scatterguard,
        parameters,
    )


@bpp.set_run_key_decorator(CONST.PLAN.GRIDSCAN_AND_MOVE)
@bpp.run_decorator(md={"subplan_name": CONST.PLAN.GRIDSCAN_AND_MOVE})
def run_gridscan_and_fetch_results(
    fgs_composite: HyperionFlyScanXRayCentreComposite,
    parameters: HyperionSpecifiedThreeDGridScan,
    feature_controlled: _FeatureControlled,
) -> MsgGenerator:
    """A multi-run plan which runs a gridscan, gets the results from zocalo
    and fires an event with the centres of mass determined by zocalo"""

    yield from feature_controlled.setup_trigger(fgs_composite, parameters)

    LOGGER.info("Starting grid scan")
    yield from bps.stage(
        fgs_composite.zocalo, group=ZOCALO_STAGE_GROUP
    )  # connect to zocalo and make sure the queue is clear
    yield from run_gridscan(fgs_composite, parameters, feature_controlled)

    LOGGER.info("Grid scan finished, getting results.")

    try:
        with TRACER.start_span("wait_for_zocalo"):
            yield from bps.trigger_and_read(
                [fgs_composite.zocalo], name=ZOCALO_READING_PLAN_NAME
            )
            LOGGER.info("Zocalo triggered and read, interpreting results.")
            xrc_results = yield from get_full_processing_results(fgs_composite.zocalo)
            LOGGER.info(f"Got xray centres, top 5: {xrc_results[:5]}")
            filtered_results = [
                result
                for result in xrc_results
                if result["total_count"] >= ZOCALO_MIN_TOTAL_COUNT_THRESHOLD
            ]
            discarded_count = len(xrc_results) - len(filtered_results)
            if discarded_count > 0:
                LOGGER.info(
                    f"Removed {discarded_count} results because below threshold"
                )
            if filtered_results:
                flyscan_results = [
                    _xrc_result_in_boxes_to_result_in_mm(xr, parameters)
                    for xr in filtered_results
                ]
            else:
                LOGGER.warning("No X-ray centre received")
                raise CrystalNotFoundException()
            yield from _fire_xray_centre_result_event(flyscan_results)

    finally:
        # Turn off dev/shm streaming to avoid filling disk, see https://github.com/DiamondLightSource/hyperion/issues/1395
        LOGGER.info("Turning off Eiger dev/shm streaming")
        yield from bps.abs_set(fgs_composite.eiger.odin.fan.dev_shm_enable, 0)  # type: ignore # Fix types in ophyd-async (https://github.com/DiamondLightSource/mx-bluesky/issues/855)

        # Wait on everything before returning to GDA (particularly apertures), can be removed
        # when we do not return to GDA here
        yield from bps.wait()


def _xrc_result_in_boxes_to_result_in_mm(
    xrc_result: XrcResult, parameters: HyperionSpecifiedThreeDGridScan
) -> XRayCentreResult:
    fgs_params = parameters.FGS_params
    xray_centre = fgs_params.grid_position_to_motor_position(
        np.array(xrc_result["centre_of_mass"])
    )
    # A correction is applied to the bounding box to map discrete grid coordinates to
    # the corners of the box in motor-space; we do not apply this correction
    # to the xray-centre as it is already in continuous space and the conversion has
    # been performed already
    # In other words, xrc_result["bounding_box"] contains the position of the box centre,
    # so we subtract half a box to get the corner of the box
    return XRayCentreResult(
        centre_of_mass_mm=xray_centre,
        bounding_box_mm=(
            fgs_params.grid_position_to_motor_position(
                np.array(xrc_result["bounding_box"][0]) - 0.5
            ),
            fgs_params.grid_position_to_motor_position(
                np.array(xrc_result["bounding_box"][1]) - 0.5
            ),
        ),
        max_count=xrc_result["max_count"],
        total_count=xrc_result["total_count"],
    )


@bpp.set_run_key_decorator(CONST.PLAN.FLYSCAN_RESULTS)
def _fire_xray_centre_result_event(results: Sequence[XRayCentreResult]):
    def empty_plan():
        return iter([])

    yield from bpp.run_wrapper(
        empty_plan(),
        md={CONST.PLAN.FLYSCAN_RESULTS: [dataclasses.asdict(r) for r in results]},
    )


@bpp.set_run_key_decorator(CONST.PLAN.GRIDSCAN_MAIN)
@bpp.run_decorator(md={"subplan_name": CONST.PLAN.GRIDSCAN_MAIN})
def run_gridscan(
    fgs_composite: HyperionFlyScanXRayCentreComposite,
    parameters: HyperionSpecifiedThreeDGridScan,
    feature_controlled: _FeatureControlled,
    md={  # noqa
        "plan_name": CONST.PLAN.GRIDSCAN_MAIN,
    },
):
    # Currently gridscan only works for omega 0, see #
    with TRACER.start_span("moving_omega_to_0"):
        yield from bps.abs_set(fgs_composite.smargon.omega, 0)

    # We only subscribe to the communicator callback for run_gridscan, so this is where
    # we should generate an event reading the values which need to be included in the
    # ispyb deposition
    with TRACER.start_span("ispyb_hardware_readings"):
        yield from standard_read_hardware_pre_collection(
            fgs_composite.undulator,
            fgs_composite.synchrotron,
            fgs_composite.s4_slit_gaps,
            fgs_composite.dcm,
            fgs_composite.smargon,
        )

    read_during_collection = partial(
        standard_read_hardware_during_collection,
        fgs_composite.aperture_scatterguard,
        fgs_composite.attenuator,
        fgs_composite.flux,
        fgs_composite.dcm,
        fgs_composite.eiger,
    )

    LOGGER.info("Setting fgs params")
    yield from feature_controlled.set_flyscan_params()

    LOGGER.info("Waiting for gridscan validity check")
    yield from wait_for_gridscan_valid(feature_controlled.fgs_motors)

    LOGGER.info("Waiting for arming to finish")
    yield from bps.wait(CONST.WAIT.GRID_READY_FOR_DC)
    yield from bps.stage(fgs_composite.eiger)

    yield from kickoff_and_complete_gridscan(
        feature_controlled.fgs_motors,
        fgs_composite.eiger,
        fgs_composite.synchrotron,
        [parameters.scan_points_first_grid, parameters.scan_points_second_grid],
        parameters.scan_indices,
        plan_during_collection=read_during_collection,
    )
    yield from bps.abs_set(feature_controlled.fgs_motors.z_steps, 0, wait=False)


def wait_for_gridscan_valid(fgs_motors: FastGridScanCommon, timeout=0.5):
    LOGGER.info("Waiting for valid fgs_params")
    SLEEP_PER_CHECK = 0.1
    times_to_check = int(timeout / SLEEP_PER_CHECK)
    for _ in range(times_to_check):
        scan_invalid = yield from bps.rd(fgs_motors.scan_invalid)
        pos_counter = yield from bps.rd(fgs_motors.position_counter)
        LOGGER.debug(
            f"Scan invalid: {scan_invalid} and position counter: {pos_counter}"
        )
        if not scan_invalid and pos_counter == 0:
            LOGGER.info("Gridscan scan valid and position counter reset")
            return
        yield from bps.sleep(SLEEP_PER_CHECK)
    raise SampleException("Scan invalid - pin too long/short/bent and out of range")


@dataclasses.dataclass
class _FeatureControlled:
    class _ZebraSetup(Protocol):
        def __call__(
            self, zebra: Zebra, group="setup_zebra_for_gridscan", wait=True
        ) -> MsgGenerator: ...

    class _ExtraSetup(Protocol):
        def __call__(
            self,
            fgs_composite: HyperionFlyScanXRayCentreComposite,
            parameters: HyperionSpecifiedThreeDGridScan,
        ) -> MsgGenerator: ...

    setup_trigger: _ExtraSetup
    tidy_plan: Callable[[HyperionFlyScanXRayCentreComposite], MsgGenerator]
    set_flyscan_params: Callable[[], MsgGenerator]
    fgs_motors: FastGridScanCommon


def _get_feature_controlled(
    fgs_composite: HyperionFlyScanXRayCentreComposite,
    parameters: HyperionSpecifiedThreeDGridScan,
):
    if parameters.features.use_panda_for_gridscan:
        return _FeatureControlled(
            setup_trigger=_panda_triggering_setup,
            tidy_plan=_panda_tidy,
            set_flyscan_params=partial(
                set_flyscan_params,
                fgs_composite.panda_fast_grid_scan,
                parameters.panda_FGS_params,
            ),
            fgs_motors=fgs_composite.panda_fast_grid_scan,
        )
    else:
        return _FeatureControlled(
            setup_trigger=_zebra_triggering_setup,
            tidy_plan=partial(_generic_tidy, group="flyscan_zebra_tidy", wait=True),
            set_flyscan_params=partial(
                set_flyscan_params,
                fgs_composite.zebra_fast_grid_scan,
                parameters.FGS_params,
            ),
            fgs_motors=fgs_composite.zebra_fast_grid_scan,
        )


def _generic_tidy(
    fgs_composite: HyperionFlyScanXRayCentreComposite, group, wait=True
) -> MsgGenerator:
    LOGGER.info("Tidying up Zebra")
    yield from tidy_up_zebra_after_gridscan(
        fgs_composite.zebra, fgs_composite.sample_shutter, group=group, wait=wait
    )
    LOGGER.info("Tidying up Zocalo")
    # make sure we don't consume any other results
    yield from bps.unstage(fgs_composite.zocalo, group=group, wait=wait)


def _panda_tidy(fgs_composite: HyperionFlyScanXRayCentreComposite):
    group = "panda_flyscan_tidy"
    LOGGER.info("Disabling panda blocks")
    yield from disarm_panda_for_gridscan(fgs_composite.panda, group)
    yield from _generic_tidy(fgs_composite, group, False)
    yield from bps.wait(group, timeout=10)
    yield from bps.unstage(fgs_composite.panda)


def _zebra_triggering_setup(
    fgs_composite: HyperionFlyScanXRayCentreComposite,
    parameters: HyperionSpecifiedThreeDGridScan,
):
    yield from setup_zebra_for_gridscan(
        fgs_composite.zebra, fgs_composite.sample_shutter, wait=True
    )


def _panda_triggering_setup(
    fgs_composite: HyperionFlyScanXRayCentreComposite,
    parameters: HyperionSpecifiedThreeDGridScan,
):
    LOGGER.info("Setting up Panda for flyscan")

    run_up_distance_mm = yield from bps.rd(
        fgs_composite.panda_fast_grid_scan.run_up_distance_mm
    )

    time_between_x_steps_ms = (
        HardwareConstants.PANDA_FGS_EIGER_DEADTIME_S + parameters.exposure_time_s
    ) * 1e3

    smargon_speed_limit_mm_per_s = yield from bps.rd(
        fgs_composite.smargon.x.max_velocity
    )

    sample_velocity_mm_per_s = (
        parameters.panda_FGS_params.x_step_size_mm * 1e3 / time_between_x_steps_ms
    )
    if sample_velocity_mm_per_s > smargon_speed_limit_mm_per_s:
        raise SmargonSpeedException(
            f"Smargon speed was calculated from x step size\
            {parameters.panda_FGS_params.x_step_size_mm}mm and\
            time_between_x_steps_ms {time_between_x_steps_ms} as\
            {sample_velocity_mm_per_s}mm/s. The smargon's speed limit is\
            {smargon_speed_limit_mm_per_s}mm/s."
        )
    else:
        LOGGER.info(
            f"Panda grid scan: Smargon speed set to {sample_velocity_mm_per_s} mm/s"
            f" and using a run-up distance of {run_up_distance_mm}"
        )

    yield from bps.mv(
        fgs_composite.panda_fast_grid_scan.time_between_x_steps_ms,
        time_between_x_steps_ms,
    )

    directory_provider_root = Path(parameters.storage_directory)
    yield from set_panda_directory(directory_provider_root)

    yield from setup_panda_for_flyscan(
        fgs_composite.panda,
        parameters.panda_FGS_params,
        fgs_composite.smargon,
        parameters.exposure_time_s,
        time_between_x_steps_ms,
        sample_velocity_mm_per_s,
    )

    LOGGER.info("Setting up Zebra for panda flyscan")
    yield from setup_zebra_for_panda_flyscan(
        fgs_composite.zebra, fgs_composite.sample_shutter, wait=True
    )



from opentelemetry import metrics, trace
from opentelemetry.exporter.otlp.proto.http.metric_exporter import OTLPMetricExporter
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor


def setup_tracing(service_name: str = "Hyperion"):
    resource = Resource(attributes={SERVICE_NAME: service_name})

    traceProvider = TracerProvider(resource=resource)
    processor = BatchSpanProcessor(
        OTLPSpanExporter(endpoint="http://0.0.0.0:4318/v1/traces")
    )
    traceProvider.add_span_processor(processor)
    trace.set_tracer_provider(traceProvider)

    reader = PeriodicExportingMetricReader(
        OTLPMetricExporter(endpoint="http://0.0.0.0:4318/v1/metrics")
    )
    meterProvider = MeterProvider(resource=resource, metric_readers=[reader])
    metrics.set_meter_provider(meterProvider)


TRACER = trace.get_tracer(__name__)



from functools import cache

from daq_config_server.client import ConfigServer
from pydantic import model_validator

from mx_bluesky.common.external_interaction.config_server import FeatureFlags
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.parameters.constants import CONST


class HyperionFeatureFlags(FeatureFlags):
    """
    Feature flags specific to Hyperion.

    Attributes:
        use_panda_for_gridscan:         If True then the PandA is used for gridscans, otherwise the zebra is used
        compare_cpu_and_gpu_zocalo:     If True then GPU result processing is enabled
            alongside CPU and the results are compared. The CPU result is still take.n
        use_gpu_results:                If True then GPU result processing is enabled
            and the GPU result is taken.
        set_stub_offsets:               If True then set the stub offsets after moving to the crystal (ignored for
            multi-centre)
        omega_flip:                     If True then invert the smargon omega motor rotation commands with respect to
         the hyperion request.
    """

    @staticmethod
    @cache
    def get_config_server() -> ConfigServer:
        return ConfigServer(CONST.CONFIG_SERVER_URL, LOGGER)

    @model_validator(mode="after")
    def use_gpu_and_compare_cannot_both_be_true(self):
        assert not (self.use_gpu_results and self.compare_cpu_and_gpu_zocalo), (
            "Cannot both use GPU results and compare them to CPU"
        )
        return self

    use_panda_for_gridscan: bool = CONST.I03.USE_PANDA_FOR_GRIDSCAN
    compare_cpu_and_gpu_zocalo: bool = CONST.I03.COMPARE_CPU_AND_GPU_ZOCALO
    use_gpu_results: bool = CONST.I03.USE_GPU_RESULTS
    set_stub_offsets: bool = CONST.I03.SET_STUB_OFFSETS
    omega_flip: bool = CONST.I03.OMEGA_FLIP



"""
Fixed target data collection
"""

from datetime import datetime
from pathlib import Path

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.attenuator.attenuator import ReadOnlyAttenuator
from dodal.devices.hutch_shutter import HutchShutter, ShutterDemand
from dodal.devices.i24.aperture import Aperture
from dodal.devices.i24.beam_center import DetectorBeamCenter
from dodal.devices.i24.beamstop import Beamstop
from dodal.devices.i24.dcm import DCM
from dodal.devices.i24.dual_backlight import DualBacklight
from dodal.devices.i24.focus_mirrors import FocusMirrorsMode
from dodal.devices.i24.i24_detector_motion import DetectorMotion
from dodal.devices.i24.pmac import PMAC
from dodal.devices.zebra.zebra import Zebra

from mx_bluesky.beamlines.i24.serial.dcid import (
    DCID,
    get_pilatus_filename_template_from_device,
    read_beam_info_from_hardware,
)
from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import (
    ChipType,
    MappingType,
    PumpProbeSetting,
)
from mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1 import (
    read_parameters,
    upload_chip_map_to_geobrick,
)
from mx_bluesky.beamlines.i24.serial.log import SSX_LOGGER, log_on_entry
from mx_bluesky.beamlines.i24.serial.parameters import FixedTargetParameters
from mx_bluesky.beamlines.i24.serial.parameters.constants import BEAM_CENTER_LUT_FILES
from mx_bluesky.beamlines.i24.serial.setup_beamline import caget, cagetstring, caput, pv
from mx_bluesky.beamlines.i24.serial.setup_beamline import setup_beamline as sup
from mx_bluesky.beamlines.i24.serial.setup_beamline.setup_zebra_plans import (
    SHUTTER_OPEN_TIME,
    arm_zebra,
    close_fast_shutter,
    open_fast_shutter,
    open_fast_shutter_at_each_position_plan,
    reset_zebra_when_collection_done_plan,
    setup_zebra_for_fastchip_plan,
)
from mx_bluesky.beamlines.i24.serial.write_nexus import call_nexgen

# Move this in common place as part of
# https://github.com/DiamondLightSource/mx-bluesky/pull/603
PMAC_MOVE_TIME = 0.008  # Move time between positions on chip ~ 7-8 ms


def calculate_collection_timeout(parameters: FixedTargetParameters) -> float:
    """Give an estimation of the time the plan should wait for the data collection \
        to be finished.

    For non-pump probe collections and collection with short delays, it should be \
    enough to use the collection time plus a genereous 30s buffer.
    For EAVA (Excite and visit again) collections instead, the laser dwell and laser \
    delay times should be included in the calculation. For long dalays between pump \
    and probe, the shutter opening time will also need to be taken into account.
    For more details on the dynamics see
    https://confluence.diamond.ac.uk/display/MXTech/Dynamics+and+fixed+targets.

    Args:
        parameters (FixedTargerParameters): The collection parameters.

    Returns:
        The estimated collection time, in s.
    """
    buffer = PMAC_MOVE_TIME * parameters.total_num_images + 600
    pump_setting = parameters.pump_repeat
    collection_time = parameters.total_num_images * parameters.exposure_time_s
    if pump_setting in [
        PumpProbeSetting.NoPP,
        PumpProbeSetting.Short1,
        PumpProbeSetting.Short2,
    ]:
        timeout = collection_time + buffer
    else:
        # EAVA: Excite and visit again
        num_windows = parameters.total_num_images / parameters.num_exposures
        timeout = (
            collection_time
            + parameters.laser_dwell_s * num_windows  # type: ignore
            + parameters.laser_delay_s
            + buffer
        )
        if pump_setting == PumpProbeSetting.Medium1:
            # Long delay between pump and probe, with fast shutter opening and closing.
            timeout = timeout + SHUTTER_OPEN_TIME * parameters.total_num_images
    return timeout


def write_userlog(
    parameters: FixedTargetParameters,
    filename: str,
    transmission: float,
    wavelength: float,
):
    # Write a record of what was collected to the processing directory
    userlog_path = Path(parameters.visit) / f"processing/{parameters.directory}"
    userlog_fid = f"{filename}_parameters.txt"
    SSX_LOGGER.debug(f"Write a user log in {userlog_path}")

    userlog_path.mkdir(parents=True, exist_ok=True)

    text = f"""
        Fixed Target Data Collection Parameters\n
        Data directory \t{parameters.collection_directory.as_posix()}\n
        Filename \t{filename}\n
        Shots per pos \t{parameters.num_exposures}\n
        Total N images \t{parameters.total_num_images}\n
        Exposure time \t{parameters.exposure_time_s}\n
        Det distance \t{parameters.detector_distance_mm}\n
        Transmission \t{transmission}\n
        Wavelength \t{wavelength}\n
        Detector type \t{parameters.detector_name}\n
        Pump status \t{parameters.pump_repeat}\n
        Pump exp time \t{parameters.laser_dwell_s}\n
        Pump delay \t{parameters.laser_delay_s}\n
    """
    with open(userlog_path / userlog_fid, "w") as f:
        f.write(text)


@log_on_entry
def get_chip_prog_values(
    parameters: FixedTargetParameters,
):
    # this is where p variables for fast laser expts will be set
    if parameters.pump_repeat in [
        PumpProbeSetting.NoPP,
        PumpProbeSetting.Short1,
        PumpProbeSetting.Short2,
        PumpProbeSetting.Medium1,
    ]:
        pump_repeat_pvar = 0
    elif parameters.pump_repeat == PumpProbeSetting.Repeat1:
        pump_repeat_pvar = 1
    elif parameters.pump_repeat == PumpProbeSetting.Repeat2:
        pump_repeat_pvar = 2
    elif parameters.pump_repeat == PumpProbeSetting.Repeat3:
        pump_repeat_pvar = 3
    elif parameters.pump_repeat == PumpProbeSetting.Repeat5:
        pump_repeat_pvar = 5
    elif parameters.pump_repeat == PumpProbeSetting.Repeat10:
        pump_repeat_pvar = 10
    else:
        raise ValueError(f"Unknown pump_repeat, pump_repeat = {parameters.pump_repeat}")

    SSX_LOGGER.info(
        f"Pump repeat is {str(parameters.pump_repeat)}, PVAR set to {pump_repeat_pvar}"
    )

    if parameters.pump_repeat == PumpProbeSetting.Short2:
        pump_in_probe = 1
    else:
        pump_in_probe = 0

    SSX_LOGGER.info(f"pump_in_probe set to {pump_in_probe}")

    chip_dict: dict[str, list] = {
        "X_NUM_STEPS": [11, parameters.chip.x_num_steps],
        "Y_NUM_STEPS": [12, parameters.chip.y_num_steps],
        "X_STEP_SIZE": [13, parameters.chip.x_step_size],
        "Y_STEP_SIZE": [14, parameters.chip.y_step_size],
        "DWELL_TIME": [15, parameters.exposure_time_s],
        "X_START": [16, 0],
        "Y_START": [17, 0],
        "Z_START": [18, 0],
        "X_NUM_BLOCKS": [20, parameters.chip.x_blocks],
        "Y_NUM_BLOCKS": [21, parameters.chip.y_blocks],
        "X_BLOCK_SIZE": [24, parameters.chip.x_block_size],
        "Y_BLOCK_SIZE": [25, parameters.chip.y_block_size],
        "COLTYPE": [26, 41],
        "N_EXPOSURES": [30, parameters.num_exposures],
        "PUMP_REPEAT": [32, pump_repeat_pvar],
        "LASER_DWELL": [34, parameters.laser_dwell_s],
        "LASERTWO_DWELL": [35, parameters.pre_pump_exposure_s],
        "LASER_DELAY": [37, parameters.laser_delay_s],
        "PUMP_IN_PROBE": [38, pump_in_probe],
    }

    chip_dict["DWELL_TIME"][1] = 1000 * parameters.exposure_time_s
    chip_dict["LASER_DWELL"][1] = (
        1000 * parameters.laser_dwell_s if parameters.laser_dwell_s else 0
    )
    chip_dict["LASERTWO_DWELL"][1] = (
        1000 * parameters.pre_pump_exposure_s if parameters.pre_pump_exposure_s else 0
    )
    chip_dict["LASER_DELAY"][1] = (
        1000 * parameters.laser_delay_s if parameters.laser_delay_s else 0
    )

    return chip_dict


@log_on_entry
def load_motion_program_data(
    pmac: PMAC,
    motion_program_dict: dict[str, list],
    map_type: int,
    pump_repeat: int,
    checker_pattern: bool,
):
    SSX_LOGGER.info("Loading motion program data for chip.")
    SSX_LOGGER.info(f"Pump_repeat is {PumpProbeSetting(pump_repeat)}")
    if pump_repeat == PumpProbeSetting.NoPP:
        if map_type == MappingType.NoMap:
            prefix = 11
            SSX_LOGGER.info(f"Map type is None, setting program prefix to {prefix}")
        elif map_type == MappingType.Lite:
            prefix = 12
        else:
            SSX_LOGGER.warning(f"Unknown Map Type, map_type = {map_type}")
            return
    elif pump_repeat in [pp.value for pp in PumpProbeSetting if pp != 0]:
        # Pump setting chosen
        prefix = 14
        SSX_LOGGER.info(f"Setting program prefix to {prefix}")
        yield from bps.abs_set(pmac.pmac_string, "P1439=0", wait=True)
        if checker_pattern:
            SSX_LOGGER.info("Checker pattern setting enabled.")
            yield from bps.abs_set(pmac.pmac_string, "P1439=1", wait=True)
        if pump_repeat == PumpProbeSetting.Medium1:
            # Medium1 has time delays (Fast shutter opening time in ms)
            yield from bps.abs_set(pmac.pmac_string, "P1441=50", wait=True)
        else:
            yield from bps.abs_set(pmac.pmac_string, "P1441=0", wait=True)
    else:
        SSX_LOGGER.warning(f"Unknown Pump repeat, pump_repeat = {pump_repeat}")
        return

    SSX_LOGGER.info("Set PMAC_STRING pv.")
    for key in sorted(motion_program_dict.keys()):
        v = motion_program_dict[key]
        pvar_base = prefix * 100
        pvar = pvar_base + v[0]
        value = str(v[1])
        s = f"P{pvar}={value}"
        SSX_LOGGER.info(f"{key} \t {s}")
        yield from bps.abs_set(pmac.pmac_string, s, wait=True)
        yield from bps.sleep(0.02)
    yield from bps.sleep(0.2)


@log_on_entry
def get_prog_num(
    chip_type: ChipType, map_type: MappingType, pump_repeat: PumpProbeSetting
) -> int:
    """Get the motion program number based on the experiment parameters set by \
    the user.
    Any pump probe experiment will return program number 14 (assumes lite mapping).
    For non pump probe experiments, the program number depends on the chip and map type:
        - Custom, Mini and PSI chips, as well as Oxford chips with no map return 11
        - Oxford chips with lite mapping return 12
        - Oxford chips with full mapping should return 13. Currently disabled, will \
            raise an error.
    """
    SSX_LOGGER.info("Get Program Number for the motion program.")
    SSX_LOGGER.info(f"Pump_repeat: {str(pump_repeat)} \t Chip Type: {str(chip_type)}")
    if pump_repeat != PumpProbeSetting.NoPP:
        SSX_LOGGER.info("Assuming Map type = Mapping Lite.")
        SSX_LOGGER.info("Program number: 14")
        return 14

    if chip_type not in [ChipType.Oxford, ChipType.OxfordInner]:
        SSX_LOGGER.info("Program number: 11")
        return 11

    if map_type == MappingType.NoMap:
        SSX_LOGGER.info(f"Map type: {str(map_type)}")
        SSX_LOGGER.info("Program number: 11")
        return 11
    if map_type == MappingType.Lite:
        SSX_LOGGER.info(f"Map type: {str(map_type)}")
        SSX_LOGGER.info("Program number: 12")
        return 12


@log_on_entry
def set_datasize(
    parameters: FixedTargetParameters,
):
    SSX_LOGGER.info(
        f"Setting PV to calculated total number of images: {parameters.total_num_images}"
    )

    SSX_LOGGER.debug(f"Map type: {parameters.map_type}")
    SSX_LOGGER.debug(f"Chip type: {parameters.chip.chip_type}")
    if parameters.map_type == MappingType.Lite:
        SSX_LOGGER.debug(f"Num exposures: {parameters.num_exposures}")
        SSX_LOGGER.debug(f"Block count: {len(parameters.chip_map)}")

    caput(pv.me14e_gp10, parameters.total_num_images)


@log_on_entry
def start_i24(
    zebra: Zebra,
    aperture: Aperture,
    backlight: DualBacklight,
    beamstop: Beamstop,
    detector_stage: DetectorMotion,
    shutter: HutchShutter,
    parameters: FixedTargetParameters,
    dcm: DCM,
    mirrors: FocusMirrorsMode,
    beam_center_device: DetectorBeamCenter,
    dcid: DCID,
):
    """Set up for I24 fixed target data collection, trigger the detector and open \
    the hutch shutter.
    Returns the start_time.
    """

    beam_settings = yield from read_beam_info_from_hardware(
        dcm, mirrors, beam_center_device, parameters.detector_name
    )
    SSX_LOGGER.info("Start I24 data collection.")
    start_time = datetime.now()
    SSX_LOGGER.info(f"Collection start time {start_time.ctime()}")

    SSX_LOGGER.debug("Set up beamline")
    yield from sup.setup_beamline_for_collection_plan(
        aperture, backlight, beamstop, wait=True
    )

    yield from sup.move_detector_stage_to_position_plan(
        detector_stage, parameters.detector_distance_mm
    )

    SSX_LOGGER.debug("Set up beamline DONE")

    filepath = parameters.collection_directory.as_posix()
    filename = parameters.filename

    SSX_LOGGER.debug("Acquire Region")

    num_gates = parameters.total_num_images // parameters.num_exposures

    SSX_LOGGER.info(f"Total number of images: {parameters.total_num_images}")
    SSX_LOGGER.info(f"Number of exposures: {parameters.num_exposures}")
    SSX_LOGGER.info(f"Number of gates (=Total images/N exposures): {num_gates:.4f}")

    if parameters.detector_name == "pilatus":
        SSX_LOGGER.info("Using Pilatus detector")
        SSX_LOGGER.info(f"Fastchip Pilatus setup: filepath {filepath}")
        SSX_LOGGER.info(f"Fastchip Pilatus setup: filename {filename}")
        SSX_LOGGER.info(
            f"Fastchip Pilatus setup: number of images {parameters.total_num_images}"
        )
        SSX_LOGGER.info(
            f"Fastchip Pilatus setup: exposure time {parameters.exposure_time_s}"
        )

        yield from sup.pilatus(
            "fastchip",
            [
                filepath,
                filename,
                parameters.total_num_images,
                parameters.exposure_time_s,
            ],
        )

        # DCID process depends on detector PVs being set up already
        SSX_LOGGER.debug("Start DCID process")
        filetemplate = yield from get_pilatus_filename_template_from_device()
        dcid.generate_dcid(
            beam_settings=beam_settings,
            image_dir=filepath,
            file_template=filetemplate,
            num_images=parameters.total_num_images,
            shots_per_position=parameters.num_exposures,
            start_time=start_time,
            pump_probe=bool(parameters.pump_repeat),
        )

        SSX_LOGGER.debug("Arm Pilatus. Arm Zebra.")
        shutter_time_offset = (
            SHUTTER_OPEN_TIME
            if parameters.pump_repeat is PumpProbeSetting.Medium1
            else 0.0
        )
        yield from setup_zebra_for_fastchip_plan(
            zebra,
            parameters.detector_name,
            num_gates,
            parameters.num_exposures,
            parameters.exposure_time_s,
            shutter_time_offset,
            wait=True,
        )
        if parameters.pump_repeat == PumpProbeSetting.Medium1:
            yield from open_fast_shutter_at_each_position_plan(
                zebra, parameters.num_exposures, parameters.exposure_time_s
            )
        caput(pv.pilat_acquire, "1")  # Arm pilatus
        yield from arm_zebra(zebra)
        caput(pv.pilat_filename, filename)
        yield from bps.sleep(1.5)

    elif parameters.detector_name == "eiger":
        SSX_LOGGER.info("Using Eiger detector")

        SSX_LOGGER.debug(f"Creating the directory for the collection in {filepath}.")

        SSX_LOGGER.info(f"Triggered Eiger setup: filepath {filepath}")
        SSX_LOGGER.info(f"Triggered Eiger setup: filename {filename}")
        SSX_LOGGER.info(
            f"Triggered Eiger setup: number of images {parameters.total_num_images}"
        )
        SSX_LOGGER.info(
            f"Triggered Eiger setup: exposure time {parameters.exposure_time_s}"
        )

        yield from sup.eiger(
            "triggered",
            [
                filepath,
                filename,
                parameters.total_num_images,
                parameters.exposure_time_s,
            ],
        )

        # DCID process depends on detector PVs being set up already
        SSX_LOGGER.debug("Start DCID process")
        complete_filename = cagetstring(pv.eiger_ODfilenameRBV)
        filetemplate = f"{complete_filename}.nxs"
        dcid.generate_dcid(
            beam_settings=beam_settings,
            image_dir=filepath,
            file_template=filetemplate,
            num_images=parameters.total_num_images,
            shots_per_position=parameters.num_exposures,
            start_time=start_time,
            pump_probe=bool(parameters.pump_repeat),
        )

        SSX_LOGGER.debug("Arm Zebra.")
        shutter_time_offset = (
            SHUTTER_OPEN_TIME
            if parameters.pump_repeat is PumpProbeSetting.Medium1
            else 0.0
        )
        yield from setup_zebra_for_fastchip_plan(
            zebra,
            parameters.detector_name,
            num_gates,
            parameters.num_exposures,
            parameters.exposure_time_s,
            shutter_time_offset,
            wait=True,
        )
        if parameters.pump_repeat == PumpProbeSetting.Medium1:
            yield from open_fast_shutter_at_each_position_plan(
                zebra, parameters.num_exposures, parameters.exposure_time_s
            )
        yield from arm_zebra(zebra)

        yield from bps.sleep(1.5)

    else:
        msg = f"Unknown Detector Type, det_type = {parameters.detector_name}"
        SSX_LOGGER.error(msg)
        raise ValueError(msg)

    # Open the hutch shutter
    yield from bps.abs_set(shutter, ShutterDemand.OPEN, wait=True)

    return start_time


@log_on_entry
def finish_i24(
    zebra: Zebra,
    pmac: PMAC,
    shutter: HutchShutter,
    dcm: DCM,
    parameters: FixedTargetParameters,
):
    SSX_LOGGER.info(
        f"Finish I24 data collection with {parameters.detector_name} detector."
    )

    complete_filename: str
    transmission = float(caget(pv.pilat_filtertrasm))
    wavelength = yield from bps.rd(dcm.wavelength_in_a)

    if parameters.detector_name == "pilatus":
        SSX_LOGGER.debug("Finish I24 Pilatus")
        complete_filename = f"{parameters.filename}_{caget(pv.pilat_filenum)}"
        yield from reset_zebra_when_collection_done_plan(zebra)
        yield from sup.pilatus("return-to-normal", None)
        yield from bps.sleep(0.2)
    elif parameters.detector_name == "eiger":
        SSX_LOGGER.debug("Finish I24 Eiger")
        yield from reset_zebra_when_collection_done_plan(zebra)
        yield from sup.eiger("return-to-normal", None)
        complete_filename = cagetstring(pv.eiger_ODfilenameRBV)  # type: ignore
    else:
        raise ValueError(f"{parameters.detector_name=} unrecognised")

    # Detector independent moves
    SSX_LOGGER.info("Move chip back to home position by setting PMAC_STRING pv.")
    yield from bps.trigger(pmac.to_xyz_zero)
    SSX_LOGGER.info("Closing shutter")
    yield from bps.abs_set(shutter, ShutterDemand.CLOSE, wait=True)

    # Write a record of what was collected to the processing directory
    write_userlog(parameters, complete_filename, transmission, wavelength)


def run_aborted_plan(pmac: PMAC, dcid: DCID):
    """Plan to send pmac_strings to tell the PMAC when a collection has been aborted, \
        either by pressing the Abort button or because of a timeout, and to reset the \
        P variable.
    """
    SSX_LOGGER.warning("Data Collection Aborted")
    yield from bps.trigger(pmac.abort_program, wait=True)

    end_time = datetime.now()
    dcid.collection_complete(end_time, aborted=True)


@log_on_entry
def main_fixed_target_plan(
    zebra: Zebra,
    pmac: PMAC,
    aperture: Aperture,
    backlight: DualBacklight,
    beamstop: Beamstop,
    detector_stage: DetectorMotion,
    shutter: HutchShutter,
    dcm: DCM,
    mirrors: FocusMirrorsMode,
    beam_center_device: DetectorBeamCenter,
    parameters: FixedTargetParameters,
    dcid: DCID,
) -> MsgGenerator:
    SSX_LOGGER.info("Running a chip collection on I24")

    beam_center_pixels = sup.compute_beam_center_position_from_lut(
        BEAM_CENTER_LUT_FILES[parameters.detector_name],
        parameters.detector_distance_mm,
        parameters.detector_size_constants,
    )
    yield from sup.set_detector_beam_center_plan(
        beam_center_device,
        beam_center_pixels,
    )

    SSX_LOGGER.info("Getting Program Dictionary")

    # If alignment type is Oxford inner it is still an Oxford type chip
    if parameters.chip.chip_type == ChipType.OxfordInner:
        SSX_LOGGER.debug("Change chip type Oxford Inner to Oxford.")
        parameters.chip.chip_type = ChipType.Oxford

    chip_prog_dict = get_chip_prog_values(parameters)
    SSX_LOGGER.info("Loading Motion Program Data")
    yield from load_motion_program_data(
        pmac,
        chip_prog_dict,
        parameters.map_type,
        parameters.pump_repeat,
        parameters.checker_pattern,
    )

    set_datasize(parameters)

    start_time = yield from start_i24(
        zebra,
        aperture,
        backlight,
        beamstop,
        detector_stage,
        shutter,
        parameters,
        dcm,
        mirrors,
        beam_center_device,
        dcid,
    )

    SSX_LOGGER.info("Moving to Start")
    yield from bps.trigger(pmac.to_xyz_zero)
    yield from bps.sleep(2.0)

    # Now ready for data collection. Open fast shutter (zebra gate)
    SSX_LOGGER.info("Opening fast shutter.")
    yield from open_fast_shutter(zebra)

    # Kick off the StartOfCollect script
    SSX_LOGGER.debug("Notify DCID of the start of the collection.")
    dcid.notify_start()

    if parameters.detector_name == "eiger":
        wavelength = yield from bps.rd(dcm.wavelength_in_a)
        beam_x = yield from bps.rd(beam_center_device.beam_x)
        beam_y = yield from bps.rd(beam_center_device.beam_y)
        SSX_LOGGER.debug("Start nexus writing service.")
        call_nexgen(
            chip_prog_dict, parameters, wavelength, (beam_x, beam_y), start_time
        )

    yield from kickoff_and_complete_collection(pmac, parameters)


def kickoff_and_complete_collection(pmac: PMAC, parameters: FixedTargetParameters):
    prog_num = get_prog_num(
        parameters.chip.chip_type, parameters.map_type, parameters.pump_repeat
    )
    yield from bps.abs_set(pmac.program_number, prog_num, group="setup_pmac")
    # Calculate approx collection time
    total_collection_time = calculate_collection_timeout(parameters)
    SSX_LOGGER.info(f"Estimated collection time: {total_collection_time}s.")
    yield from bps.abs_set(
        pmac.collection_time, total_collection_time, group="setup_pmac"
    )
    yield from bps.wait(group="setup_pmac")  # Make sure the soft signals are set

    @bpp.run_decorator(md={"subplan_name": "run_ft_collection"})
    def run_collection():
        SSX_LOGGER.info(f"Kick off PMAC with program number {prog_num}.")
        yield from bps.kickoff(pmac.run_program, wait=True)
        yield from bps.complete(pmac.run_program, wait=True)
        SSX_LOGGER.info("Collection completed without errors.")

    yield from run_collection()


@log_on_entry
def collection_complete_plan(
    dcid: DCID, collection_directory: Path, map_type: MappingType
) -> MsgGenerator:
    end_time = datetime.now()
    SSX_LOGGER.debug(f"Collection end time {end_time}")
    dcid.collection_complete(end_time, aborted=False)

    # NOTE no files to copy anymore but shoud write userlog here
    yield from bps.null()


@log_on_entry
def tidy_up_after_collection_plan(
    zebra: Zebra,
    pmac: PMAC,
    shutter: HutchShutter,
    dcm: DCM,
    parameters: FixedTargetParameters,
    dcid: DCID,
) -> MsgGenerator:
    """A plan to be run to tidy things up at the end af a fixed target collection, \
    both successful or aborted.
    """
    SSX_LOGGER.info("Closing fast shutter")
    yield from close_fast_shutter(zebra)
    yield from bps.sleep(2.0)

    # This probably should go in main then
    if parameters.detector_name == "pilatus":
        SSX_LOGGER.debug("Pilatus Acquire STOP")
        caput(pv.pilat_acquire, 0)
    elif parameters.detector_name == "eiger":
        SSX_LOGGER.debug("Eiger Acquire STOP")
        caput(pv.eiger_acquire, 0)
        caput(pv.eiger_ODcapture, "Done")
    yield from bps.sleep(0.5)

    yield from finish_i24(zebra, pmac, shutter, dcm, parameters)

    SSX_LOGGER.debug("Notify DCID of end of collection.")
    dcid.notify_end()

    SSX_LOGGER.debug("Quick summary of settings")
    SSX_LOGGER.debug(
        f"Chip name = {parameters.filename} sub_dir = {parameters.directory}"
    )


def run_fixed_target_plan(
    zebra: Zebra = inject("zebra"),
    pmac: PMAC = inject("pmac"),
    aperture: Aperture = inject("aperture"),
    backlight: DualBacklight = inject("backlight"),
    beamstop: Beamstop = inject("beamstop"),
    detector_stage: DetectorMotion = inject("detector_motion"),
    shutter: HutchShutter = inject("shutter"),
    dcm: DCM = inject("dcm"),
    mirrors: FocusMirrorsMode = inject("focus_mirrors"),
    attenuator: ReadOnlyAttenuator = inject("attenuator"),
) -> MsgGenerator:
    # Read the parameters
    parameters: FixedTargetParameters = yield from read_parameters(
        detector_stage, attenuator
    )

    # Create collection directory
    parameters.collection_directory.mkdir(parents=True, exist_ok=True)

    if parameters.chip_map:
        yield from upload_chip_map_to_geobrick(pmac, parameters.chip_map)

    beam_center_device = sup.get_beam_center_device(parameters.detector_name)

    # DCID instance - do not create yet
    dcid = DCID(emit_errors=False, expt_params=parameters)

    yield from bpp.contingency_wrapper(
        main_fixed_target_plan(
            zebra,
            pmac,
            aperture,
            backlight,
            beamstop,
            detector_stage,
            shutter,
            dcm,
            mirrors,
            beam_center_device,
            parameters,
            dcid,
        ),
        except_plan=lambda e: (yield from run_aborted_plan(pmac, dcid)),
        final_plan=lambda: (
            yield from tidy_up_after_collection_plan(
                zebra, pmac, shutter, dcm, parameters, dcid
            )
        ),
        auto_raise=False,
    )



from functools import partial
from unittest.mock import AsyncMock, MagicMock, call, patch

import numpy as np
import pytest
from bluesky import plan_stubs as bps
from bluesky.plan_stubs import null
from bluesky.run_engine import RunEngine, RunEngineResult
from dodal.devices.backlight import Backlight
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.oav.pin_image_recognition.utils import SampleLocation
from dodal.devices.oav.utils import PinNotFoundException
from dodal.devices.smargon import Smargon
from ophyd.sim import NullStatus
from ophyd_async.testing import get_mock_put, set_mock_value

from mx_bluesky.common.utils.exceptions import SampleException, WarningException
from mx_bluesky.hyperion.device_setup_plans.smargon import (
    move_smargon_warn_on_out_of_range,
)
from mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan import (
    DEFAULT_STEP_SIZE,
    PinTipCentringComposite,
    move_pin_into_view,
    pin_tip_centre_plan,
    trigger_and_return_pin_tip,
)


def get_fake_pin_values_generator(x, y):
    yield from null()
    return np.array([x, y])


FAKE_EDGE_ARRAYS = np.array([1.0, 2.0, 3.0]), np.array([3.0, 4.0, 5.0])


@pytest.fixture
def mock_pin_tip(pin_tip: PinTipDetection):
    pin_tip._get_tip_and_edge_data = AsyncMock(return_value=pin_tip.INVALID_POSITION)
    return pin_tip


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.bps.sleep",
    new=MagicMock(),
)
async def test_given_the_pin_tip_is_already_in_view_when_get_tip_into_view_then_tip_returned_and_smargon_not_moved(
    smargon: Smargon, oav: OAV, RE: RunEngine, mock_pin_tip: PinTipDetection
):
    set_mock_value(mock_pin_tip.triggered_tip, np.array([100, 200]))

    mock_pin_tip.trigger = MagicMock(return_value=NullStatus())

    result = RE(move_pin_into_view(mock_pin_tip, smargon))

    mock_pin_tip.trigger.assert_called_once()
    assert await smargon.x.user_setpoint.get_value() == 0
    assert isinstance(result, RunEngineResult)
    assert result.plan_result == (100, 200)
    assert all(type(_) is int for _ in result.plan_result)


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.bps.sleep",
    new=MagicMock(),
)
async def test_given_no_tip_found_but_will_be_found_when_get_tip_into_view_then_smargon_moved_positive_and_tip_returned(
    smargon: Smargon, oav: OAV, RE: RunEngine, mock_pin_tip: PinTipDetection
):
    set_mock_value(mock_pin_tip.validity_timeout, 0.015)

    def set_pin_tip_when_x_moved(f, *args, **kwargs):
        mock_pin_tip._get_tip_and_edge_data.return_value = SampleLocation(  # type: ignore
            100, 200, *FAKE_EDGE_ARRAYS
        )
        return f(*args, **kwargs)

    x_user_setpoint = get_mock_put(smargon.x.user_setpoint)
    x_user_setpoint.side_effect = partial(
        set_pin_tip_when_x_moved, x_user_setpoint.side_effect
    )

    result = RE(move_pin_into_view(mock_pin_tip, smargon))

    assert await smargon.x.user_setpoint.get_value() == DEFAULT_STEP_SIZE
    assert isinstance(result, RunEngineResult)
    assert result.plan_result == (100, 200)


@pytest.mark.parametrize(
    "expected_step_size, returned_location",
    [[DEFAULT_STEP_SIZE, (None, None)], [-DEFAULT_STEP_SIZE, (0, 100)]],
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.bps.sleep",
    new=MagicMock(),
)
async def test_tip_found_only_after_all_iterations_exhausted_in_the_same_direction_then_tip_returned(
    smargon: Smargon,
    oav: OAV,
    RE: RunEngine,
    mock_pin_tip: PinTipDetection,
    expected_step_size: float,
    returned_location: tuple[int, int],
):
    set_mock_value(mock_pin_tip.validity_timeout, 0.015)
    mock_pin_tip._get_tip_and_edge_data.return_value = SampleLocation(  # type: ignore
        returned_location[0], returned_location[1], *FAKE_EDGE_ARRAYS
    )

    iterations = 0

    def set_pin_tip_when_x_moved(f, *args, **kwargs):
        nonlocal iterations
        iterations += 1
        location = returned_location
        if iterations == 2:
            location = (100, 200)

        mock_pin_tip._get_tip_and_edge_data.return_value = SampleLocation(  # type: ignore
            location[0], location[1], *FAKE_EDGE_ARRAYS
        )

        return f(*args, **kwargs)

    x_user_setpoint = get_mock_put(smargon.x.user_setpoint)
    x_user_setpoint.side_effect = partial(
        set_pin_tip_when_x_moved, x_user_setpoint.side_effect
    )

    result = RE(move_pin_into_view(mock_pin_tip, smargon, max_steps=2))

    x_user_setpoint.assert_has_calls(
        [call(expected_step_size, wait=True), call(expected_step_size * 2, wait=True)]
    )
    assert isinstance(result, RunEngineResult)
    assert result.plan_result == (100, 200)
    assert all(type(_) is int for _ in result.plan_result)


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.bps.sleep",
    new=MagicMock(),
)
async def test_given_tip_at_zero_but_will_be_found_when_get_tip_into_view_then_smargon_moved_negative_and_tip_returned(
    smargon: Smargon, oav: OAV, RE: RunEngine, mock_pin_tip: PinTipDetection
):
    mock_pin_tip._get_tip_and_edge_data.return_value = SampleLocation(  # type: ignore
        0, 100, *FAKE_EDGE_ARRAYS
    )
    set_mock_value(mock_pin_tip.validity_timeout, 0.15)

    def set_pin_tip_when_x_moved(f, *args, **kwargs):
        mock_pin_tip._get_tip_and_edge_data.return_value = SampleLocation(  # type: ignore
            100, 200, *FAKE_EDGE_ARRAYS
        )
        return f(*args, **kwargs)

    x_user_setpoint = get_mock_put(smargon.x.user_setpoint)
    x_user_setpoint.side_effect = partial(
        set_pin_tip_when_x_moved, x_user_setpoint.side_effect
    )

    result = RE(move_pin_into_view(mock_pin_tip, smargon))

    assert await smargon.x.user_setpoint.get_value() == -DEFAULT_STEP_SIZE
    assert result.plan_result == (100, 200)  # type: ignore


async def test_trigger_and_return_pin_tip_works_for_AD_pin_tip_detection(
    oav: OAV, RE: RunEngine, mock_pin_tip: PinTipDetection
):
    mock_pin_tip._get_tip_and_edge_data.return_value = SampleLocation(  # type: ignore
        200, 100, *FAKE_EDGE_ARRAYS
    )
    set_mock_value(mock_pin_tip.validity_timeout, 0.15)
    re_result = RE(trigger_and_return_pin_tip(mock_pin_tip))
    assert all(re_result.plan_result == (200, 100))  # type: ignore


def test_trigger_and_return_pin_tip_works_for_ophyd_pin_tip_detection(
    ophyd_pin_tip_detection: PinTipDetection, RE: RunEngine
):
    mock_trigger_result = SampleLocation(100, 200, np.array([]), np.array([]))
    ophyd_pin_tip_detection._get_tip_and_edge_data = AsyncMock(
        return_value=mock_trigger_result
    )
    re_result = RE(trigger_and_return_pin_tip(ophyd_pin_tip_detection))
    assert all(re_result.plan_result == (100, 200))  # type: ignore


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.trigger_and_return_pin_tip"
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.bps.sleep",
    new=MagicMock(),
)
async def test_pin_tip_starting_near_negative_edge_doesnt_exceed_limit(
    mock_trigger_and_return_tip: MagicMock,
    smargon: Smargon,
    oav: OAV,
    RE: RunEngine,
    pin_tip: PinTipDetection,
):
    mock_trigger_and_return_tip.side_effect = [
        get_fake_pin_values_generator(0, 100),
        get_fake_pin_values_generator(0, 100),
    ]

    set_mock_value(smargon.x.user_setpoint, -1.8)
    set_mock_value(smargon.x.user_readback, -1.8)

    with pytest.raises(WarningException):
        RE(move_pin_into_view(pin_tip, smargon, max_steps=1))

    assert await smargon.x.user_setpoint.get_value() == -2


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.trigger_and_return_pin_tip"
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.bps.sleep",
    new=MagicMock(),
)
async def test_pin_tip_starting_near_positive_edge_doesnt_exceed_limit(
    mock_trigger_and_return_pin_tip: MagicMock,
    smargon: Smargon,
    oav: OAV,
    RE: RunEngine,
    pin_tip: PinTipDetection,
):
    mock_trigger_and_return_pin_tip.side_effect = [
        get_fake_pin_values_generator(
            PinTipDetection.INVALID_POSITION[0], PinTipDetection.INVALID_POSITION[1]
        ),
        get_fake_pin_values_generator(
            PinTipDetection.INVALID_POSITION[0], PinTipDetection.INVALID_POSITION[1]
        ),
    ]
    set_mock_value(smargon.x.user_setpoint, 1.8)
    set_mock_value(smargon.x.user_readback, 1.8)

    with pytest.raises(WarningException):
        RE(move_pin_into_view(pin_tip, smargon, max_steps=1))

    assert await smargon.x.user_setpoint.get_value() == 2


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.bps.sleep",
    new=MagicMock(),
)
async def test_given_no_tip_found_ever_when_get_tip_into_view_then_smargon_moved_positive_and_exception_thrown(
    smargon: Smargon, oav: OAV, RE: RunEngine, pin_tip: PinTipDetection
):
    set_mock_value(pin_tip.triggered_tip, pin_tip.INVALID_POSITION)
    set_mock_value(pin_tip.validity_timeout, 0.01)

    with pytest.raises(WarningException):
        RE(move_pin_into_view(pin_tip, smargon))

    assert await smargon.x.user_setpoint.get_value() == 1


def test_given_moving_out_of_range_when_move_with_warn_called_then_warning_exception(
    RE: RunEngine, smargon: Smargon
):
    set_mock_value(smargon.x.high_limit_travel, 10)

    with pytest.raises(WarningException):
        RE(move_smargon_warn_on_out_of_range(smargon, (100, 0, 0)))


def return_pixel(pixel, *args):
    yield from null()
    return pixel


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.wait_for_tip_to_be_found",
    new=partial(return_pixel, (200, 200)),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.get_move_required_so_that_beam_is_at_pixel",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.move_pin_into_view",
    new=partial(return_pixel, (100, 100)),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.pre_centring_setup_oav",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.bps.sleep",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.move_smargon_warn_on_out_of_range",
    autospec=True,
)
async def test_when_pin_tip_centre_plan_called_then_expected_plans_called(
    move_smargon,
    mock_sleep,
    mock_setup_oav,
    get_move: MagicMock,
    smargon: Smargon,
    oav: OAV,
    test_config_files: dict[str, str],
    RE: RunEngine,
):
    set_mock_value(oav.zoom_controller.level, "1.0")
    composite = PinTipCentringComposite(
        backlight=MagicMock(spec=Backlight),
        oav=oav,
        smargon=smargon,
        pin_tip_detection=MagicMock(spec=PinTipDetection),
    )
    RE(pin_tip_centre_plan(composite, 50, test_config_files["oav_config_json"]))

    assert mock_setup_oav.call_count == 1

    assert len(get_move.call_args_list) == 2

    args, _ = get_move.call_args_list[0]
    assert args[1] == (117, 100)

    assert await smargon.omega.user_setpoint.get_value() == 90

    args, _ = get_move.call_args_list[1]
    assert args[1] == (217, 200)


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.wait_for_tip_to_be_found",
    new=partial(return_pixel, (200, 200)),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.get_move_required_so_that_beam_is_at_pixel",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.move_pin_into_view",
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.pre_centring_setup_oav",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.bps.sleep",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.move_smargon_warn_on_out_of_range",
    autospec=True,
)
def test_given_pin_tip_detect_using_ophyd_when_pin_tip_centre_plan_called_then_expected_plans_called(
    move_smargon,
    mock_sleep,
    mock_setup_oav,
    mock_move_into_view,
    get_move: MagicMock,
    smargon: Smargon,
    oav: OAV,
    test_config_files: dict[str, str],
    RE: RunEngine,
):
    set_mock_value(smargon.omega.user_readback, 0)
    mock_ophyd_pin_tip_detection = MagicMock(spec=PinTipDetection)
    composite = PinTipCentringComposite(
        backlight=MagicMock(Backlight),
        oav=oav,
        smargon=smargon,
        pin_tip_detection=mock_ophyd_pin_tip_detection,
    )
    mock_move_into_view.side_effect = partial(return_pixel, (100, 100))
    RE(pin_tip_centre_plan(composite, 50, test_config_files["oav_config_json"]))

    mock_move_into_view.assert_called_once_with(mock_ophyd_pin_tip_detection, smargon)

    assert mock_setup_oav.call_count == 1


@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.get_move_required_so_that_beam_is_at_pixel",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.move_pin_into_view",
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.pre_centring_setup_oav",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.bps.sleep",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.move_smargon_warn_on_out_of_range",
    autospec=True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.pin_tip_centring_plan.wait_for_tip_to_be_found",
    autospec=True,
)
def test_warning_raised_if_pin_tip_goes_out_of_view_after_rotation(
    mock_wait_for_tip,
    move_smargon,
    mock_sleep,
    mock_setup_oav,
    mock_move_into_view,
    get_move: MagicMock,
    smargon: Smargon,
    oav: OAV,
    test_config_files: dict[str, str],
    RE: RunEngine,
):
    set_mock_value(smargon.omega.user_readback, 0)
    mock_ophyd_pin_tip_detection = MagicMock(spec=PinTipDetection)
    composite = PinTipCentringComposite(
        backlight=MagicMock(Backlight),
        oav=oav,
        smargon=smargon,
        pin_tip_detection=mock_ophyd_pin_tip_detection,
    )

    def raise_exception(*args):
        yield from bps.null()
        raise PinNotFoundException()

    mock_wait_for_tip.side_effect = raise_exception
    mock_move_into_view.side_effect = partial(return_pixel, (100, 100))
    with pytest.raises(SampleException):
        RE(pin_tip_centre_plan(composite, 50, test_config_files["oav_config_json"]))
    assert mock_wait_for_tip.call_count == 1



import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.beamlines import aithre
from dodal.devices.aithre_lasershaping.goniometer import Goniometer

from mx_bluesky.beamlines.aithre_lasershaping import (
    change_goniometer_turn_speed,
    rotate_goniometer_relative,
)


@pytest.fixture
def goniometer(RE: RunEngine) -> Goniometer:
    return aithre.goniometer(connect_immediately=True, mock=True)


def test_goniometer_relative_rotation(
    sim_run_engine: RunEngineSimulator, goniometer: Goniometer
):
    msgs = sim_run_engine.simulate_plan(rotate_goniometer_relative(15, goniometer))
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "goniometer-omega"
        and msg.args[0] == 15,
    )


def test_change_goniometer_turn_speed(
    sim_run_engine: RunEngineSimulator, goniometer: Goniometer
):
    msgs = sim_run_engine.simulate_plan(change_goniometer_turn_speed(40, goniometer))
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "goniometer-omega-velocity"
        and msg.args[0] == 40,
    )
